"""Open Responses streaming event models.

Maps PSE state events to Open Responses SSE events.
"""

from typing import Literal

from pydantic import BaseModel, Field

from orchard.server.models.responses.output import (
    IncompleteDetails,
    OutputFunctionCall,
    OutputMessage,
    OutputReasoning,
    OutputStatus,
    OutputTextContent,
    ReasoningContent,
    ResponseUsage,
    generate_message_id,
    generate_response_id,
    get_current_timestamp,
)
from orchard.server.models.tools import generate_tool_call_id

# --- Response Lifecycle Events ---


class ResponseCreatedEvent(BaseModel):
    """Emitted when response generation begins."""

    type: Literal["response.created"] = "response.created"
    sequence_number: int
    response: "ResponseSnapshot"


class ResponseInProgressEvent(BaseModel):
    """Emitted when response generation is actively in progress."""

    type: Literal["response.in_progress"] = "response.in_progress"
    sequence_number: int
    response: "ResponseSnapshot"


class ResponseCompletedEvent(BaseModel):
    """Emitted when response generation completes successfully."""

    type: Literal["response.completed"] = "response.completed"
    sequence_number: int
    response: "ResponseSnapshot"


class ResponseFailedEvent(BaseModel):
    """Emitted when response generation fails."""

    type: Literal["response.failed"] = "response.failed"
    sequence_number: int
    response: "ResponseSnapshot"


class ResponseIncompleteEvent(BaseModel):
    """Emitted when response generation stops before completion."""

    type: Literal["response.incomplete"] = "response.incomplete"
    sequence_number: int
    response: "ResponseSnapshot"


# --- Output Item Events ---


class OutputItemAddedEvent(BaseModel):
    """Emitted when a new output item starts."""

    type: Literal["response.output_item.added"] = "response.output_item.added"
    sequence_number: int
    output_index: int
    item: OutputMessage | OutputFunctionCall | OutputReasoning


class OutputItemDoneEvent(BaseModel):
    """Emitted when an output item is complete."""

    type: Literal["response.output_item.done"] = "response.output_item.done"
    sequence_number: int
    output_index: int
    item: OutputMessage | OutputFunctionCall | OutputReasoning


# --- Content Part Events ---


class ContentPartAddedEvent(BaseModel):
    """Emitted when a new content part starts within an item."""

    type: Literal["response.content_part.added"] = "response.content_part.added"
    sequence_number: int
    item_id: str
    output_index: int
    content_index: int
    part: OutputTextContent | ReasoningContent


class ContentPartDoneEvent(BaseModel):
    """Emitted when a content part is complete."""

    type: Literal["response.content_part.done"] = "response.content_part.done"
    sequence_number: int
    item_id: str
    output_index: int
    content_index: int
    part: OutputTextContent | ReasoningContent


# --- Text Delta Events ---


class OutputTextDeltaEvent(BaseModel):
    """Emitted for each text chunk in an assistant message."""

    type: Literal["response.output_text.delta"] = "response.output_text.delta"
    sequence_number: int
    item_id: str
    output_index: int
    content_index: int
    delta: str
    logprobs: list = Field(default_factory=list)


class OutputTextDoneEvent(BaseModel):
    """Emitted when text generation for a content part is complete."""

    type: Literal["response.output_text.done"] = "response.output_text.done"
    sequence_number: int
    item_id: str
    output_index: int
    content_index: int
    text: str
    logprobs: list = Field(default_factory=list)


# --- Function Call Argument Events ---


class FunctionCallArgumentsDeltaEvent(BaseModel):
    """Emitted for each chunk of function call arguments."""

    type: Literal["response.function_call_arguments.delta"] = (
        "response.function_call_arguments.delta"
    )
    sequence_number: int
    item_id: str
    output_index: int
    delta: str


class FunctionCallArgumentsDoneEvent(BaseModel):
    """Emitted when function call arguments are complete."""

    type: Literal["response.function_call_arguments.done"] = (
        "response.function_call_arguments.done"
    )
    sequence_number: int
    item_id: str
    output_index: int
    arguments: str


# --- Reasoning Events ---


class ReasoningDeltaEvent(BaseModel):
    """Emitted for each chunk of reasoning content."""

    type: Literal["response.reasoning.delta"] = "response.reasoning.delta"
    sequence_number: int
    item_id: str
    output_index: int
    content_index: int
    delta: str


class ReasoningDoneEvent(BaseModel):
    """Emitted when reasoning content is complete."""

    type: Literal["response.reasoning.done"] = "response.reasoning.done"
    sequence_number: int
    item_id: str
    output_index: int
    content_index: int
    text: str


class ReasoningSummaryTextDeltaEvent(BaseModel):
    """Emitted for each chunk of reasoning summary."""

    type: Literal["response.reasoning_summary_text.delta"] = (
        "response.reasoning_summary_text.delta"
    )
    sequence_number: int
    item_id: str
    output_index: int
    summary_index: int
    delta: str


class ReasoningSummaryTextDoneEvent(BaseModel):
    """Emitted when reasoning summary is complete."""

    type: Literal["response.reasoning_summary_text.done"] = (
        "response.reasoning_summary_text.done"
    )
    sequence_number: int
    item_id: str
    output_index: int
    summary_index: int
    text: str


# --- Error Event ---


class StreamErrorEvent(BaseModel):
    """Emitted when an error occurs during streaming."""

    type: Literal["error"] = "error"
    sequence_number: int
    error: dict


# --- Response Snapshot (partial state during streaming) ---


class ResponseSnapshot(BaseModel):
    """Snapshot of response state at a point in the stream."""

    id: str = Field(default_factory=generate_response_id)
    object: Literal["response"] = "response"
    created_at: int = Field(default_factory=get_current_timestamp)
    completed_at: int | None = None
    status: OutputStatus = OutputStatus.IN_PROGRESS
    incomplete_details: IncompleteDetails | None = None
    model: str
    output: list[OutputMessage | OutputFunctionCall | OutputReasoning] = Field(
        default_factory=list
    )
    usage: ResponseUsage | None = None


# --- Stream State Tracker ---


class StreamingOutputItem:
    """Tracks state for a single output item during streaming."""

    def __init__(
        self,
        item_id: str,
        output_index: int,
        item_type: str,
        call_id: str | None = None,
        function_name: str | None = None,
    ):
        self.item_id = item_id
        self.output_index = output_index
        self.item_type = item_type
        self.call_id = call_id
        self.function_name = function_name
        self.accumulated_content = ""
        self.accumulated_arguments = ""
        self.content_index = 0
        self.status = OutputStatus.IN_PROGRESS

    def to_skeleton(self) -> OutputMessage | OutputFunctionCall | OutputReasoning:
        """Create the skeleton item for output_item.added event."""
        if self.item_type == "message":
            return OutputMessage(
                id=self.item_id,
                status=OutputStatus.IN_PROGRESS,
                content=[],
            )
        elif self.item_type == "tool_call":
            return OutputFunctionCall(
                id=self.item_id,
                call_id=self.call_id or generate_tool_call_id(),
                name=self.function_name or "",
                arguments="",
                status=OutputStatus.IN_PROGRESS,
            )
        else:  # reasoning
            return OutputReasoning(
                id=self.item_id,
                status=OutputStatus.IN_PROGRESS,
                content=[],
                summary=[],
            )

    def to_completed(self) -> OutputMessage | OutputFunctionCall | OutputReasoning:
        """Create the completed item for output_item.done event."""
        if self.item_type == "message":
            return OutputMessage(
                id=self.item_id,
                status=OutputStatus.COMPLETED,
                content=[OutputTextContent(text=self.accumulated_content)],
            )
        elif self.item_type == "tool_call":
            return OutputFunctionCall(
                id=self.item_id,
                call_id=self.call_id or generate_tool_call_id(),
                name=self.function_name or "",
                arguments=self.accumulated_arguments,
                status=OutputStatus.COMPLETED,
            )
        else:  # reasoning
            return OutputReasoning(
                id=self.item_id,
                status=OutputStatus.COMPLETED,
                content=[ReasoningContent(text=self.accumulated_content)],
                summary=[],
            )


class ResponseStreamState:
    """Tracks the complete state of a response during streaming."""

    def __init__(self, response_id: str, model: str):
        self.response_id = response_id
        self.model = model
        self.created_at = get_current_timestamp()
        self.completed_at: int | None = None
        self.items: dict[int, StreamingOutputItem] = {}  # output_index -> item
        self.sequence_number = 0
        self.status = OutputStatus.IN_PROGRESS
        self.incomplete_details: IncompleteDetails | None = None
        self.usage: ResponseUsage | None = None

    def next_sequence_number(self) -> int:
        """Get and increment the sequence number."""
        seq = self.sequence_number
        self.sequence_number += 1
        return seq

    def get_or_create_item(
        self,
        output_index: int,
        item_type: str,
        identifier: str | None = None,
    ) -> StreamingOutputItem:
        """Get existing item or create new one for the given output_index."""
        if output_index not in self.items:
            item_id = generate_message_id()
            call_id = None
            function_name = None

            if item_type == "tool_call" and identifier:
                call_id = generate_tool_call_id()
                function_name = identifier.removeprefix("tool_call:")

            self.items[output_index] = StreamingOutputItem(
                item_id=item_id,
                output_index=output_index,
                item_type=item_type,
                call_id=call_id,
                function_name=function_name,
            )
        return self.items[output_index]

    def snapshot(self) -> ResponseSnapshot:
        """Create a snapshot of current response state."""
        output_items = [
            item.to_completed()
            if item.status == OutputStatus.COMPLETED
            else item.to_skeleton()
            for item in sorted(self.items.values(), key=lambda x: x.output_index)
        ]
        return ResponseSnapshot(
            id=self.response_id,
            created_at=self.created_at,
            completed_at=self.completed_at,
            status=self.status,
            incomplete_details=self.incomplete_details,
            model=self.model,
            output=output_items,
            usage=self.usage,
        )


# Union of all streaming event types
StreamingEvent = (
    ResponseCreatedEvent
    | ResponseInProgressEvent
    | ResponseCompletedEvent
    | ResponseFailedEvent
    | ResponseIncompleteEvent
    | OutputItemAddedEvent
    | OutputItemDoneEvent
    | ContentPartAddedEvent
    | ContentPartDoneEvent
    | OutputTextDeltaEvent
    | OutputTextDoneEvent
    | FunctionCallArgumentsDeltaEvent
    | FunctionCallArgumentsDoneEvent
    | ReasoningDeltaEvent
    | ReasoningDoneEvent
    | ReasoningSummaryTextDeltaEvent
    | ReasoningSummaryTextDoneEvent
    | StreamErrorEvent
)



import asyncio
import builtins
import json
import logging
import random
from collections.abc import AsyncIterable
from contextlib import AsyncExitStack
from typing import Any

from fastapi import APIRouter, HTTPException, status
from fastapi.responses import JSONResponse
from sse_starlette.sse import EventSourceResponse

from orchard.formatter.multimodal import (
    build_multimodal_layout,
    build_multimodal_messages,
)
from orchard.ipc.serialization import _build_request_payload
from orchard.ipc.utils import (
    ResponseDeltaDict,
    release_delta_resources,
)
from orchard.server.dependencies import IPCStateDep, ModelRegistryDep
from orchard.server.exceptions import InferenceError
from orchard.server.models.reasoning import normalize_reasoning_value
from orchard.server.models.responses import (
    ContentPartAddedEvent,
    ContentPartDoneEvent,
    FunctionCallArgumentsDeltaEvent,
    FunctionCallArgumentsDoneEvent,
    IncompleteDetails,
    InputTokensDetails,
    OutputFunctionCall,
    OutputItemAddedEvent,
    OutputItemDoneEvent,
    OutputMessage,
    OutputReasoning,
    OutputStatus,
    OutputTextContent,
    OutputTextDeltaEvent,
    OutputTextDoneEvent,
    OutputTokensDetails,
    ReasoningContent,
    ReasoningDeltaEvent,
    ReasoningDoneEvent,
    ResponseCompletedEvent,
    ResponseCreatedEvent,
    ResponseFailedEvent,
    ResponseIncompleteEvent,
    ResponseInProgressEvent,
    ResponseObject,
    ResponseRequest,
    ResponseStreamState,
    ResponseUsage,
    StreamErrorEvent,
    generate_response_id,
    get_current_timestamp,
)
from orchard.server.routes._common import (
    _ModelNotReadyError,
    extract_usage,
    managed_stream_session,
    resolve_model,
)

logger = logging.getLogger(__name__)

responses_router = APIRouter()


@responses_router.post(
    "/responses",
    response_model=None,  # Disable automatic response model to support streaming
    summary="Create a model response",
    tags=["Responses"],
)
async def handle_response_request(
    request: ResponseRequest,
    ipc_state: IPCStateDep,
    model_registry: ModelRegistryDep,
) -> ResponseObject | EventSourceResponse | JSONResponse:
    """Handle multimodal requests to the `/v1/responses` endpoint."""
    logger.info("Handling response request for model: %s", request.model)

    try:
        canonical_id, model_info = await resolve_model(model_registry, request.model)
    except _ModelNotReadyError as exc:
        return exc.response

    formatter = model_info.formatter

    try:
        messages_for_template, image_buffers, capabilities, content_order = (
            build_multimodal_messages(
                formatter=formatter,
                items=request.get_message_items(),
                instructions=request.instructions,
            )
        )
    except (ValueError, TypeError) as exc:
        logger.error("Invalid multimodal payload for request: %s", exc)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=str(exc),
        ) from exc

    if not messages_for_template:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Response request must include at least one content segment.",
        )

    # Tool definitions — used for template rendering and PSE grammar compilation
    tools_payload = (
        [tool.to_dict() for tool in request.tools] if request.tools else None
    )
    tool_schemas_json = json.dumps(tools_payload) if tools_payload else ""

    try:
        prompt_text = formatter.apply_template(
            messages_for_template,
            reasoning=request.reasoning is not None,
            tools=tools_payload,
        )
    except Exception as exc:  # pragma: no cover - defensive
        logger.exception("Failed to render chat template: %s", exc)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Failed to render chat template.",
        ) from exc

    try:
        layout_segments = build_multimodal_layout(
            prompt_text,
            image_buffers,
            capabilities,
            content_order,
            formatter.image_placeholder,
            formatter.should_clip_image_placeholder,
        )
    except ValueError as exc:
        logger.error("Failed to build multimodal layout: %s", exc)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail=str(exc)
        ) from exc

    if formatter.should_clip_image_placeholder:
        prompt_text = prompt_text.replace(formatter.image_placeholder, "")

    logger.info("Prompt text: %s", prompt_text)

    current_request_id = await ipc_state.get_next_request_id()
    logger.debug(
        "Generated request ID %d for responses submission.", current_request_id
    )
    response_channel_id = ipc_state.response_channel_id or current_request_id

    temperature = request.temperature if request.temperature is not None else 1.0
    top_p = request.top_p if request.top_p is not None else 1.0
    top_k = request.top_k if request.top_k is not None else -1
    min_p = request.min_p if request.min_p is not None else 0.0
    max_output_tokens = request.max_output_tokens or 0
    rng_seed = random.randint(0, 2**32 - 1)

    response_format_json = json.dumps(request.text.to_dict()) if request.text else ""
    reasoning_effort = normalize_reasoning_value(request.reasoning)

    response_queue: asyncio.Queue[ResponseDeltaDict] = asyncio.Queue()
    exit_stack = AsyncExitStack()
    await exit_stack.enter_async_context(
        managed_stream_session(ipc_state, current_request_id, response_queue)
    )

    try:
        prompt_bytes = prompt_text.encode("utf-8")
        capabilities_payload = [
            {"name": cap.name, "payload": cap.payload, "position": 0}
            for cap in capabilities
        ]
        prompt_payload = {
            "prompt_bytes": prompt_bytes,
            "image_buffers": image_buffers,
            "capabilities": capabilities_payload,
            "layout": layout_segments,
            "sampling_params": {
                "temperature": temperature,
                "top_p": top_p,
                "top_k": top_k,
                "min_p": min_p,
                "rng_seed": rng_seed,
            },
            "logits_params": {
                "top_logprobs": request.top_logprobs or 0,
                "frequency_penalty": request.frequency_penalty or 0.0,
                "logit_bias": {},
                "presence_penalty": request.presence_penalty or 0.0,
                "repetition_context_size": 60,
                "repetition_penalty": 1.0,
            },
            "max_generated_tokens": max_output_tokens,
            "stop_sequences": [],
            "tool_schemas_json": tool_schemas_json,
            "response_format_json": response_format_json,
            "num_candidates": 1,
            "best_of": 1,
            "final_candidates": 1,
            "task_name": request.task,
            "reasoning_effort": reasoning_effort,
            "max_tool_calls": request.max_tool_calls,
            "tool_calling_tokens": formatter.get_tool_calling_tokens(),
            "tool_choice": request.tool_choice.to_dict()
            if request.tool_choice
            else "auto",
        }
        request_bytes = _build_request_payload(
            request_id=current_request_id,
            model_id=canonical_id,
            model_path=model_info.model_path,
            request_type="generation",
            response_channel_id=response_channel_id,
            prompts=[prompt_payload],
        )

        socket = ipc_state.request_socket
        if socket is None:
            raise RuntimeError("Request socket is not initialized.")

        await socket.asend(request_bytes)
        logger.info(
            "Submitted responses request %d with %d layout segments (%d images).",
            current_request_id,
            len(layout_segments),
            len(image_buffers),
        )

        if request.stream:
            logger.debug(
                "Starting streaming response for request %d", current_request_id
            )
            response_id = generate_response_id()

            async def event_stream() -> AsyncIterable[dict[str, str]]:
                try:
                    async for event in stream_response_generator(
                        request_id=current_request_id,
                        queue=response_queue,
                        response_id=response_id,
                        model_name=request.model,
                    ):
                        yield event
                finally:
                    await exit_stack.aclose()

            return EventSourceResponse(
                content=event_stream(),
                media_type="text/event-stream",
            )

        # Non-streaming path
        aggregated = await gather_non_streaming_response(
            current_request_id, response_queue
        )

        usage = ResponseUsage(
            input_tokens=aggregated["prompt_tokens"],
            output_tokens=aggregated["completion_tokens"],
            total_tokens=aggregated["total_tokens"],
            input_tokens_details=InputTokensDetails(
                cached_tokens=aggregated["cached_tokens"],
            ),
            output_tokens_details=OutputTokensDetails(
                reasoning_tokens=aggregated["reasoning_tokens"],
            ),
        )

        incomplete_details = aggregated.get("incomplete_details")
        response = ResponseObject(
            model=request.model,
            output=aggregated["output"],
            usage=usage,
            status=OutputStatus.INCOMPLETE
            if incomplete_details
            else OutputStatus.COMPLETED,
            completed_at=aggregated.get("completed_at"),
            incomplete_details=incomplete_details,
            metadata=request.metadata,
            min_p=request.min_p,
            top_p=request.top_p,
            top_k=request.top_k,
            temperature=request.temperature,
            presence_penalty=request.presence_penalty,
            frequency_penalty=request.frequency_penalty,
            truncation=request.truncation,
            parallel_tool_calls=request.parallel_tool_calls or False,
            tool_choice=request.tool_choice,
            tools=request.tools or [],
            max_tool_calls=request.max_tool_calls,
            text=request.text,
        )

        logger.info("Response request %d completed successfully.", current_request_id)
        await exit_stack.aclose()
        return response
    except HTTPException:
        await exit_stack.aclose()
        raise
    except InferenceError as exc:
        await exit_stack.aclose()
        logger.error(
            "Inference error during response request %d: %s",
            current_request_id,
            exc,
        )
        raise HTTPException(
            status_code=status.HTTP_502_BAD_GATEWAY,
            detail=str(exc),
        ) from exc
    except Exception as exc:  # pragma: no cover - defensive
        await exit_stack.aclose()
        logger.exception(
            "Failed to process multimodal response request %d: %s",
            current_request_id,
            exc,
        )
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="An unexpected error occurred during response generation.",
        ) from exc


async def gather_non_streaming_response(
    request_id: int,
    queue: asyncio.Queue[ResponseDeltaDict],
) -> dict[str, Any]:
    """Aggregate non-streaming deltas into a final response with proper output items."""

    # Track output items by output_index
    output_items: dict[int, dict[str, Any]] = {}
    usage_counts = {
        "prompt_tokens": 0,
        "completion_tokens": 0,
        "total_tokens": 0,
        "cached_tokens": 0,
        "reasoning_tokens": 0,
    }
    error_detail: str | None = None
    finish_reason: str | None = None
    completed_at: int | None = None

    while True:
        try:
            delta = await asyncio.wait_for(queue.get(), timeout=30.0)
        except builtins.TimeoutError as exc:
            logger.error(
                "Timeout waiting for response delta for request %d", request_id
            )
            raise InferenceError("Timeout receiving response from engine.") from exc
        else:
            try:
                if not delta:
                    logger.debug(
                        "Received empty delta for response request %d", request_id
                    )
                    continue

                if delta.get("error"):
                    error_detail = str(delta["error"])
                else:
                    status_value = str(delta.get("status", "")).lower()
                    finish_value = str(delta.get("finish_reason", "")).lower()
                    if status_value == "error" or finish_value == "error":
                        error_detail = str(
                            delta.get("content")
                            or delta.get("message")
                            or delta.get("error")
                            or "Response generation failed."
                        )

                # Process state_events from PIE
                state_events = delta.get("state_events", [])
                for event in state_events:
                    _process_state_event_for_output(event, output_items)

                # Extract usage from delta
                extract_usage(delta, usage_counts)

                # Track finish reason for incomplete_details
                if fr := delta.get("finish_reason"):
                    finish_reason = str(fr).lower()

                if delta.get("is_final_delta", False):
                    completed_at = get_current_timestamp()
                    logger.debug(
                        "Received final delta for response request %d",
                        request_id,
                    )
                    break
            finally:
                queue.task_done()
                if delta:
                    release_delta_resources(delta)

    if error_detail:
        logger.error(
            "Error reported in response stream for request %d: %s",
            request_id,
            error_detail,
        )
        raise InferenceError(error_detail)

    if usage_counts["total_tokens"] <= 0:
        usage_counts["total_tokens"] = (
            usage_counts["prompt_tokens"] + usage_counts["completion_tokens"]
        )

    # Build final output items
    output = _build_output_items(output_items)

    # Determine if response was incomplete
    incomplete_details = None
    if finish_reason in ("length", "max_tokens", "max_output_tokens"):
        incomplete_details = IncompleteDetails(reason="max_output_tokens")
    elif finish_reason == "content_filter":
        incomplete_details = IncompleteDetails(reason="content_filter")

    return {
        "output": output,
        "prompt_tokens": usage_counts["prompt_tokens"],
        "completion_tokens": usage_counts["completion_tokens"],
        "total_tokens": usage_counts["total_tokens"],
        "cached_tokens": usage_counts["cached_tokens"],
        "reasoning_tokens": usage_counts["reasoning_tokens"],
        "completed_at": completed_at,
        "incomplete_details": incomplete_details,
    }


def _process_state_event_for_output(
    event: dict[str, Any],
    output_items: dict[int, dict[str, Any]],
) -> None:
    """Process a single state event and update output_items tracking."""
    event_type = event.get("event_type")
    item_type = event.get("item_type", "message")
    output_index = event.get("output_index", 0)
    identifier = event.get("identifier", "")

    if output_index not in output_items:
        output_items[output_index] = {
            "type": item_type,
            "content": "",
            "arguments": "",
            "identifier": identifier,
        }

    item = output_items[output_index]

    # Sub-item events (e.g. "arguments" within a tool call).
    if item_type == "tool_call" and identifier == "arguments":
        if event_type == "content_delta":
            item["arguments"] += event.get("delta", "")
        elif event_type == "item_completed" and "value" in event:
            item["arguments"] = str(event["value"])
        return

    # Top-level item events.
    if event_type == "content_delta":
        item["content"] += event.get("delta", "")
    elif event_type == "item_completed":
        item["completed"] = True
        if item_type == "tool_call":
            item["identifier"] = identifier
        elif "value" in event:
            item["content"] = str(event["value"])


def _build_output_items(
    output_items: dict[int, dict[str, Any]],
) -> list[OutputMessage | OutputFunctionCall | OutputReasoning]:
    """Build final output item models from accumulated state."""
    result: list[OutputMessage | OutputFunctionCall | OutputReasoning] = []

    for output_index in sorted(output_items.keys()):
        item = output_items[output_index]
        item_type = item.get("type", "message")
        content = item.get("content", "")
        identifier = item.get("identifier", "")

        if item_type == "message":
            result.append(
                OutputMessage(
                    status=OutputStatus.COMPLETED,
                    content=[OutputTextContent(text=content)] if content else [],
                )
            )
        elif item_type == "tool_call":
            function_name = identifier.removeprefix("tool_call:")
            result.append(
                OutputFunctionCall(
                    name=function_name,
                    arguments=item.get("arguments", ""),
                    status=OutputStatus.COMPLETED,
                )
            )
        elif item_type == "reasoning":
            result.append(
                OutputReasoning(
                    status=OutputStatus.COMPLETED,
                    content=[ReasoningContent(text=content)] if content else [],
                )
            )

    # If no items were built, create a default empty message
    if not result:
        result.append(OutputMessage(status=OutputStatus.COMPLETED, content=[]))

    return result


async def stream_response_generator(
    request_id: int,
    queue: asyncio.Queue[ResponseDeltaDict],
    response_id: str,
    model_name: str,
) -> AsyncIterable[dict[str, str]]:
    """Generate Open Responses SSE events from PIE deltas with state_events."""
    stream_state = ResponseStreamState(response_id=response_id, model=model_name)

    # Emit response.created
    yield _format_sse_event(
        ResponseCreatedEvent(
            sequence_number=stream_state.next_sequence_number(),
            response=stream_state.snapshot(),
        )
    )

    # Emit response.in_progress
    yield _format_sse_event(
        ResponseInProgressEvent(
            sequence_number=stream_state.next_sequence_number(),
            response=stream_state.snapshot(),
        )
    )

    error_occurred = False
    error_detail: str | None = None
    finish_reason: str | None = None

    while True:
        try:
            delta = await asyncio.wait_for(queue.get(), timeout=30.0)
        except TimeoutError:
            logger.error(
                "Timeout waiting for delta for streaming request %d", request_id
            )
            error_occurred = True
            error_detail = "Timeout receiving response from engine."
            break
        else:
            try:
                if not delta:
                    continue

                if delta.get("error"):
                    error_occurred = True
                    error_detail = str(delta["error"])
                    break

                status_value = str(delta.get("status", "")).lower()
                finish_value = str(delta.get("finish_reason", "")).lower()
                if status_value == "error" or finish_value == "error":
                    error_occurred = True
                    error_detail = str(
                        delta.get("content")
                        or delta.get("message")
                        or delta.get("error")
                        or "Response generation failed."
                    )
                    break

                # Process state_events from PIE
                state_events = delta.get("state_events") or []
                for event in state_events:
                    async for sse_event in _process_state_event_for_streaming(
                        event, stream_state
                    ):
                        yield sse_event

                # Extract usage info
                if usage := delta.get("usage"):
                    if isinstance(usage, dict):
                        stream_state.usage = ResponseUsage(
                            input_tokens=usage.get("input_tokens", 0),
                            output_tokens=usage.get("output_tokens", 0),
                            total_tokens=usage.get("total_tokens", 0),
                            input_tokens_details=InputTokensDetails(
                                cached_tokens=usage.get("cached_tokens", 0),
                            ),
                            output_tokens_details=OutputTokensDetails(
                                reasoning_tokens=usage.get("reasoning_tokens", 0),
                            ),
                        )

                # Extract token details from delta (PIE sends these at top level)
                cached_tokens = delta.get("cached_token_count")
                reasoning_tokens = delta.get("reasoning_tokens")
                if cached_tokens is not None or reasoning_tokens is not None:
                    current_usage = stream_state.usage or ResponseUsage(
                        input_tokens=0, output_tokens=0, total_tokens=0
                    )
                    stream_state.usage = ResponseUsage(
                        input_tokens=current_usage.input_tokens,
                        output_tokens=current_usage.output_tokens,
                        total_tokens=current_usage.total_tokens,
                        input_tokens_details=InputTokensDetails(
                            cached_tokens=cached_tokens or 0,
                        ),
                        output_tokens_details=OutputTokensDetails(
                            reasoning_tokens=reasoning_tokens or 0,
                        ),
                    )

                # Track finish reason for incomplete detection
                if fr := delta.get("finish_reason"):
                    finish_reason = str(fr).lower()

                if delta.get("is_final_delta", False):
                    logger.debug(
                        "Received final delta for streaming request %d", request_id
                    )
                    break
            finally:
                queue.task_done()
                if delta:
                    release_delta_resources(delta)

    # Emit completion events for items that didn't receive an item_completed event
    # from PIE. Items that did receive item_completed already have status=COMPLETED
    # (set in _process_state_event_for_streaming) and are skipped here.
    for output_index, item in sorted(stream_state.items.items()):
        if item.status != OutputStatus.COMPLETED:
            item.status = OutputStatus.COMPLETED
            # Emit done events based on item type
            if item.item_type == "message":
                yield _format_sse_event(
                    OutputTextDoneEvent(
                        sequence_number=stream_state.next_sequence_number(),
                        item_id=item.item_id,
                        output_index=output_index,
                        content_index=0,
                        text=item.accumulated_content,
                    )
                )
                yield _format_sse_event(
                    ContentPartDoneEvent(
                        sequence_number=stream_state.next_sequence_number(),
                        item_id=item.item_id,
                        output_index=output_index,
                        content_index=0,
                        part=OutputTextContent(text=item.accumulated_content),
                    )
                )
            elif item.item_type == "tool_call":
                yield _format_sse_event(
                    FunctionCallArgumentsDoneEvent(
                        sequence_number=stream_state.next_sequence_number(),
                        item_id=item.item_id,
                        output_index=output_index,
                        arguments=item.accumulated_content,
                    )
                )
            elif item.item_type == "reasoning":
                yield _format_sse_event(
                    ReasoningDoneEvent(
                        sequence_number=stream_state.next_sequence_number(),
                        item_id=item.item_id,
                        output_index=output_index,
                        content_index=0,
                        text=item.accumulated_content,
                    )
                )

            yield _format_sse_event(
                OutputItemDoneEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    output_index=output_index,
                    item=item.to_completed(),
                )
            )

    # Emit final response event
    if error_occurred:
        stream_state.status = OutputStatus.FAILED
        yield _format_sse_event(
            ResponseFailedEvent(
                sequence_number=stream_state.next_sequence_number(),
                response=stream_state.snapshot(),
            )
        )
        if error_detail:
            yield _format_sse_event(
                StreamErrorEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    error={"message": error_detail, "type": "inference_error"},
                )
            )
    else:
        # Set completed_at timestamp
        stream_state.completed_at = get_current_timestamp()

        # Determine if response was incomplete (truncated)
        is_incomplete = finish_reason in ("length", "max_tokens", "max_output_tokens")
        if is_incomplete:
            stream_state.status = OutputStatus.INCOMPLETE
            stream_state.incomplete_details = IncompleteDetails(
                reason="max_output_tokens"
            )
            yield _format_sse_event(
                ResponseIncompleteEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    response=stream_state.snapshot(),
                )
            )
        elif finish_reason == "content_filter":
            stream_state.status = OutputStatus.INCOMPLETE
            stream_state.incomplete_details = IncompleteDetails(reason="content_filter")
            yield _format_sse_event(
                ResponseIncompleteEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    response=stream_state.snapshot(),
                )
            )
        else:
            stream_state.status = OutputStatus.COMPLETED
            yield _format_sse_event(
                ResponseCompletedEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    response=stream_state.snapshot(),
                )
            )

    # Emit terminal [DONE] event (required by spec)
    yield {"data": "[DONE]"}

    logger.info("SSE stream for responses request %d completed", request_id)


async def _process_state_event_for_streaming(
    event: dict[str, Any],
    stream_state: ResponseStreamState,
) -> AsyncIterable[dict[str, str]]:
    """Process a single PSE state event and yield Open Responses SSE events."""
    event_type = event.get("event_type")
    item_type = event.get("item_type", "message")
    output_index = event.get("output_index", 0)
    identifier = event.get("identifier", "")

    item = stream_state.get_or_create_item(output_index, item_type, identifier)

    # Sub-item events (e.g. "arguments" within a tool call) are handled
    # separately from top-level item events.
    if item_type == "tool_call" and identifier == "arguments":
        delta_text = event.get("delta", "")
        if event_type == "content_delta":
            item.accumulated_arguments += delta_text
            yield _format_sse_event(
                FunctionCallArgumentsDeltaEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    delta=delta_text,
                )
            )
        elif event_type == "item_completed":
            yield _format_sse_event(
                FunctionCallArgumentsDoneEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    arguments=item.accumulated_arguments,
                )
            )
        return

    # Top-level item events.
    if event_type == "item_started":
        yield _format_sse_event(
            OutputItemAddedEvent(
                sequence_number=stream_state.next_sequence_number(),
                output_index=output_index,
                item=item.to_skeleton(),
            )
        )
        if item_type == "message":
            yield _format_sse_event(
                ContentPartAddedEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    part=OutputTextContent(text=""),
                )
            )
        elif item_type == "reasoning":
            yield _format_sse_event(
                ContentPartAddedEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    part=ReasoningContent(text=""),
                )
            )
        return

    if event_type == "content_delta":
        delta_text = event.get("delta", "")
        item.accumulated_content += delta_text
        if item_type == "message":
            yield _format_sse_event(
                OutputTextDeltaEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    delta=delta_text,
                )
            )
        elif item_type == "reasoning":
            yield _format_sse_event(
                ReasoningDeltaEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    delta=delta_text,
                )
            )
        return

    if event_type == "item_completed":
        item.status = OutputStatus.COMPLETED
        if item_type == "message":
            yield _format_sse_event(
                OutputTextDoneEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    text=item.accumulated_content,
                )
            )
            yield _format_sse_event(
                ContentPartDoneEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    part=OutputTextContent(text=item.accumulated_content),
                )
            )
        elif item_type == "tool_call":
            item.function_name = identifier.removeprefix("tool_call:")
        elif item_type == "reasoning":
            yield _format_sse_event(
                ReasoningDoneEvent(
                    sequence_number=stream_state.next_sequence_number(),
                    item_id=item.item_id,
                    output_index=output_index,
                    content_index=0,
                    text=item.accumulated_content,
                )
            )

        yield _format_sse_event(
            OutputItemDoneEvent(
                sequence_number=stream_state.next_sequence_number(),
                output_index=output_index,
                item=item.to_completed(),
            )
        )


def _format_sse_event(event: Any) -> dict[str, str]:
    """Format a streaming event for SSE transport."""
    return {
        "event": event.type,
        "data": event.model_dump_json(exclude_none=True),
    }



import json

import httpx
import pytest

pytestmark = pytest.mark.asyncio

MODEL_ID = "meta-llama/Llama-3.1-8B-Instruct"


async def test_chat_completion_multi_token_non_streaming(live_server):
    """
    Tests that the engine can correctly answer a simple question,
    verifying the RoPE offset logic is fixed.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [{"role": "user", "content": "What is the capital of France?"}],
        "temperature": 0.0,  # Use greedy sampling for a deterministic answer
        "max_completion_tokens": 10,
        "logprobs": True,
        "top_logprobs": 5,
        "stream": False,
    }

    async with httpx.AsyncClient(timeout=60.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions", json=request_payload
        )

    # --- Assertions ---
    assert response.status_code == 200, f"Expected 200 OK, got {response.status_code}"
    response_data = response.json()

    assert "id" in response_data
    assert "choices" in response_data
    assert isinstance(response_data["choices"], list)
    assert len(response_data["choices"]) > 0

    choice = response_data["choices"][0]
    assert "message" in choice
    if "content" in choice["message"] and choice["message"]["content"] is not None:
        # The model's greedy output should be "The capital of France is Paris." or something very similar.
        # We check if "Paris" is in the generated text.
        if "Paris" not in choice["message"]["content"]:
            # Pretty print logprobs for the second token
            if choice.get("logprobs"):
                logprobs_data = choice["logprobs"]
                if logprobs_data.get("content"):
                    for i, data in enumerate(logprobs_data["content"]):
                        print(f"Token {i}: {data.get('token', 'N/A')}")
                        print(f"Logprob {i}: {data.get('logprob', 'N/A')}")
                        if "top_logprobs" in data:
                            print("\nTop Logprobs:")
                            for entry in data["top_logprobs"][:5]:
                                print(
                                    f"  {entry.get('token', 'N/A'):20s} -> {entry.get('logprob', 'N/A'):.4f}"
                                )
                    print("=" * 30 + "\n")
        assert "Paris" in choice["message"]["content"], (
            f"Expected 'Paris' in response but got: '{choice['message']['content']}'"
        )
        print(choice["message"]["content"])
    else:
        pytest.fail(f"Expected 'content' in choice['message'], but got: {choice}")

    # Assert that the generation was stopped because of the token limit
    assert choice["finish_reason"].lower() in ["stop", "length"]
    assert "usage" in response_data, "Expected 'usage' field in response"
    usage = response_data["usage"]

    assert usage.get("input_tokens", 0) > 0, (
        "Expected 'input_tokens' to be greater than 0"
    )
    assert usage.get("total_tokens") == usage.get("input_tokens", 0) + usage.get(
        "output_tokens", 0
    )


async def test_chat_completion_multi_token_streaming(live_server):
    """
    Tests a multi-token, streaming chat completion request.
    Verifies that the system streams multiple tokens correctly.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [
            {"role": "user", "content": "Tell me a very short story in one sentence."}
        ],
        "max_completion_tokens": 10,
        "temperature": 0.0,
        "stream": True,
    }

    chunks_received = 0
    full_content = ""
    finish_reason = None

    async with httpx.AsyncClient(timeout=30.0) as client:
        async with client.stream(
            "POST",
            f"{server_url}/v1/chat/completions",
            json=request_payload,
        ) as response:
            assert response.status_code == 200

            async for line in response.aiter_lines():
                if line.startswith("data: "):
                    data_str = line[6:]  # Remove "data: " prefix
                    if data_str == "[DONE]":
                        break

                    chunk_data = json.loads(data_str)
                    chunks_received += 1

                    assert "id" in chunk_data
                    assert "choices" in chunk_data
                    assert len(chunk_data["choices"]) > 0

                    choice = chunk_data["choices"][0]
                    if "delta" in choice and "content" in choice["delta"]:
                        content = choice["delta"]["content"]
                        if content:
                            full_content += content

                    if "finish_reason" in choice:
                        finish_reason = choice["finish_reason"]

    # --- Assertions ---
    assert chunks_received > 1, f"Expected multiple chunks, but got {chunks_received}"
    assert len(full_content) > 0, "Expected non-empty content"
    assert finish_reason is not None, "Expected a finish reason"
    assert finish_reason.lower() in ["stop", "length"]



import asyncio
import json
import logging
import random
from collections import defaultdict
from collections.abc import AsyncIterable
from contextlib import AsyncExitStack
from typing import Any

from fastapi import APIRouter, HTTPException, status
from fastapi.responses import JSONResponse
from sse_starlette.sse import EventSourceResponse

from orchard.ipc.serialization import _build_request_payload
from orchard.ipc.utils import (
    ResponseDeltaDict,
    normalise_delta_payload,
    release_delta_resources,
)
from orchard.server.dependencies import IPCStateDep, ModelRegistryDep
from orchard.server.exceptions import InferenceError
from orchard.server.models.chat import (
    ChatCompletionChoice,
    ChatCompletionChunk,
    ChatCompletionChunkChoice,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionUsage,
    ChatMessage,
)
from orchard.server.models.chat.logprobs import ChatCompletionLogProbs
from orchard.server.models.chat.output import (
    generate_chat_completion_id,
    get_current_timestamp,
)
from orchard.server.routes._common import (
    _ModelNotReadyError,
    managed_stream_session,
    resolve_model,
)
from orchard.server.utils.batching import normalize_chat_request

logger = logging.getLogger(__name__)

chat_router = APIRouter()


def _dedupe_stop_sequences(raw_stop: list[str] | str | None) -> list[str]:
    if not raw_stop:
        return []

    if isinstance(raw_stop, str):
        candidates = [raw_stop]
    else:
        candidates = [item for item in raw_stop if item]

    seen: set[str] = set()
    unique: list[str] = []
    for seq in candidates:
        if seq not in seen:
            seen.add(seq)
            unique.append(seq)
    return unique


@chat_router.post(
    "/chat/completions",
    summary="Create a chat completion",
    tags=["Chat"],
    response_model=None,  # Disable automatic response model to support streaming
)
async def handle_completion_request(
    request: ChatCompletionRequest,
    ipc_state: IPCStateDep,
    model_registry: ModelRegistryDep,
) -> ChatCompletionResponse | EventSourceResponse | JSONResponse:
    """
    Handles requests to the `/v1/chat/completions` endpoint.
    """
    logger.info("Chat completion request for model: %s", request.model)

    try:
        canonical_id, model_info = await resolve_model(model_registry, request.model)
    except _ModelNotReadyError as exc:
        return exc.response

    formatter = model_info.formatter
    model_path = model_info.model_path

    try:
        normalized_instances = normalize_chat_request(request)
    except ValueError as exc:
        logger.error("Validation error normalizing batch parameters: %s", exc)
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=str(exc),
        ) from exc

    batch_size = len(normalized_instances)
    logger.debug("Normalized chat batch size: %d", batch_size)

    fanout_counts = [instance.best_of for instance in normalized_instances]
    final_candidate_counts = [
        instance.final_candidates for instance in normalized_instances
    ]
    total_expected_sequences = sum(fanout_counts)
    if total_expected_sequences <= 0:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="At least one candidate must be requested.",
        )

    if batch_size == 0:
        logger.error("No prompts provided.")
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="At least one prompt is required.",
        )

    prompt_payloads: list[dict[str, Any]] = []
    for instance in normalized_instances:
        messages_as_dicts = [
            msg.model_dump(exclude_none=True) for msg in instance.messages
        ]
        tools_payload = (
            [tool.to_dict() for tool in instance.tools] if instance.tools else None
        )
        tool_schemas_str = json.dumps(tools_payload) if tools_payload else ""
        prompt_text = formatter.apply_template(
            messages_as_dicts,
            reasoning=instance.reasoning_effort is not None,
            tools=tools_payload,
        )
        logger.info("Prompt text: %s", prompt_text)
        stop_sequences = _dedupe_stop_sequences(instance.stop_sequences)
        response_format_str = (
            json.dumps(instance.response_format.to_dict())
            if instance.response_format
            else ""
        )

        conversation_bytes = prompt_text.encode("utf-8")
        layout: list[dict[str, Any]] = [
            {"type": "text", "length": len(conversation_bytes)}
        ]

        payload: dict[str, Any] = {
            "prompt": prompt_text,
            "sampling_params": {
                "temperature": instance.temperature,
                "top_p": instance.top_p if instance.top_p is not None else 1.0,
                "top_k": instance.top_k if instance.top_k is not None else -1,
                "min_p": instance.min_p if instance.min_p is not None else 0.0,
                "rng_seed": random.randint(0, 2**32 - 1),
            },
            "logits_params": {
                "top_logprobs": instance.top_logprobs,
                "frequency_penalty": 0.0,
                "logit_bias": {},
                "presence_penalty": 0.0,
                "repetition_context_size": 60,
                "repetition_penalty": 1.0,
            },
            "max_generated_tokens": instance.max_completion_tokens or 1024,
            "stop_sequences": stop_sequences,
            "tool_schemas_json": tool_schemas_str,
            "response_format_json": response_format_str,
            "image_buffers": [],
            "layout": layout,
            "num_candidates": instance.best_of,
            "best_of": instance.best_of,
            "final_candidates": instance.final_candidates,
        }
        if instance.task is not None:
            payload["task_name"] = instance.task
        if instance.reasoning_effort is not None:
            payload["reasoning_effort"] = instance.reasoning_effort

        payload["tool_calling_tokens"] = formatter.get_tool_calling_tokens()
        payload["tool_choice"] = (
            request.tool_choice.to_dict() if request.tool_choice else "auto"
        )
        prompt_payloads.append(payload)

    current_request_id = await ipc_state.get_next_request_id()
    logger.debug("Generated request ID: %d", current_request_id)
    response_channel_id = ipc_state.response_channel_id or current_request_id

    response_queue = asyncio.Queue[ResponseDeltaDict]()
    exit_stack = AsyncExitStack()
    await exit_stack.enter_async_context(
        managed_stream_session(ipc_state, current_request_id, response_queue)
    )

    try:
        logger.debug("Submitting batched request %d to C++ engine", current_request_id)
        request_bytes = _build_request_payload(
            request_id=current_request_id,
            model_id=canonical_id,
            model_path=model_path,
            request_type="generation",
            response_channel_id=response_channel_id,
            prompts=prompt_payloads,
        )
        socket = ipc_state.request_socket
        if socket is None:
            raise RuntimeError("Request socket is not initialized.")

        await socket.asend(request_bytes)
        logger.info("Submitted batched request %d successfully", current_request_id)
        if request.stream:
            logger.debug("Handling streaming for request %d", current_request_id)

            async def event_stream() -> AsyncIterable[dict[str, str]]:
                try:
                    async for chunk in stream_response_generator(
                        current_request_id,
                        response_queue,
                        ipc_state,
                        request.model,
                        total_expected_sequences,
                    ):
                        yield chunk
                finally:
                    await exit_stack.aclose()

            return EventSourceResponse(
                content=event_stream(),
                media_type="text/event-stream",
            )

        logger.debug("Handling non-streaming for request %d", current_request_id)
        response_data = await gather_non_streaming_batch_response(
            current_request_id,
            response_queue,
            fanout_counts,
            final_candidate_counts,
        )
        usage = ChatCompletionUsage(
            input_tokens=response_data["prompt_tokens"],
            output_tokens=response_data["completion_tokens"],
            total_tokens=response_data["prompt_tokens"]
            + response_data["completion_tokens"],
        )
        final_response = ChatCompletionResponse(
            id=generate_chat_completion_id(),
            created=get_current_timestamp(),
            model=request.model,
            choices=response_data["choices"],
            usage=usage,
        )
        logger.info("Non-streaming request %d successful.", current_request_id)
        await exit_stack.aclose()
        return final_response
    except HTTPException:
        await exit_stack.aclose()
        raise
    except Exception as e:
        await exit_stack.aclose()
        logger.error("Error submitting request: %s", e, exc_info=True)
        raise HTTPException(
            status_code=500,
            detail="An unexpected error occurred during completion.",
        ) from e


async def gather_non_streaming_batch_response(
    request_id: int,
    queue: asyncio.Queue[ResponseDeltaDict],
    prompt_fanout_counts: list[int],
    prompt_final_counts: list[int],
) -> dict[str, Any]:
    """Collects deltas for all sequences belonging to a parent request."""

    total_expected = sum(prompt_fanout_counts)
    prompt_states: list[list[dict[str, Any]]] = [
        [
            {
                "content": "",
                "finish_reason": "unknown",
                "completion_tokens": 0,
                "logprobs_entries": [],
                "sequence_id": None,
                "completed": False,
                "error_message": None,
                "cumulative_logprob": None,
                "generation_len": 0,
            }
            for _ in range(candidate_count)
        ]
        for candidate_count in prompt_fanout_counts
    ]
    prompt_token_totals: list[int] = [0 for _ in prompt_fanout_counts]

    remaining_sequences = total_expected

    while remaining_sequences > 0:
        try:
            delta = await asyncio.wait_for(queue.get(), timeout=30.0)
        except TimeoutError as exc:
            logger.error(
                "Timeout waiting for delta for non-streaming request %d", request_id
            )
            raise InferenceError("Timeout receiving response from engine.") from exc
        else:
            try:
                normalise_delta_payload(delta)
                prompt_index = delta.get("prompt_index")
                candidate_index = delta.get("candidate_index")
                sequence_id = delta.get("sequence_id")

                if prompt_index is None or candidate_index is None:
                    logger.warning(
                        "Received delta missing prompt/candidate index for request %d: %s",
                        request_id,
                        delta,
                    )
                    continue

                if not (0 <= prompt_index < len(prompt_states)):
                    logger.warning(
                        "Prompt index %s out of range for request %d",
                        prompt_index,
                        request_id,
                    )
                    continue

                if not (0 <= candidate_index < len(prompt_states[prompt_index])):
                    logger.warning(
                        "Candidate index %s out of range for request %d prompt %d",
                        candidate_index,
                        request_id,
                        prompt_index,
                    )
                    continue

                state = prompt_states[prompt_index][candidate_index]
                if state["sequence_id"] is None and sequence_id is not None:
                    state["sequence_id"] = sequence_id

                prompt_token_count = delta.get("prompt_token_count", 0)
                if prompt_token_count > 0 and prompt_token_totals[prompt_index] == 0:
                    prompt_token_totals[prompt_index] = prompt_token_count

                delta_content = delta.get("content") or ""
                if delta_content:
                    state["content"] += delta_content

                tokens = delta.get("tokens", [])
                state["completion_tokens"] += len(tokens)

                if top_logprobs_data := delta.get("top_logprobs"):
                    top_logprobs_list: list[ChatCompletionLogProbs] = []
                    for item in top_logprobs_data:
                        token_val = (
                            (item.get("token") or "") if isinstance(item, dict) else ""
                        )
                        try:
                            raw_logprob = item.get("logprob", item.get("probability"))
                            if not isinstance(raw_logprob, (int | float)):
                                continue
                            logprob_val = round(float(raw_logprob), 6)
                        except (TypeError, ValueError):
                            logger.debug(
                                "Skipping top_logprobs entry without numeric logprob for request %d: %s",
                                request_id,
                                item,
                            )
                            continue

                        top_logprobs_list.append(
                            ChatCompletionLogProbs(
                                token=token_val,
                                logprob=logprob_val,
                                bytes=list(token_val.encode("utf-8")),
                            )
                        )

                    chosen_token_str = delta_content
                    chosen_token_logprob = -999.0
                    for item in top_logprobs_data:
                        if (
                            isinstance(item, dict)
                            and item.get("token") == chosen_token_str
                            and (
                                raw_logprob := item.get(
                                    "logprob", item.get("probability")
                                )
                            )
                            is not None
                        ):
                            if isinstance(raw_logprob, (int | float)):
                                chosen_token_logprob = float(raw_logprob)
                                break

                    logprob_entry = ChatCompletionLogProbs(
                        token=chosen_token_str,
                        logprob=chosen_token_logprob,
                        bytes=list(chosen_token_str.encode("utf-8")),
                        top_logprobs=top_logprobs_list,
                    )
                    state["logprobs_entries"].append(logprob_entry)

                finish_reason = delta.get("finish_reason") or "unknown"
                if finish_reason.lower() == "error" and delta_content:
                    state["error_message"] = delta_content

                if delta.get("is_final_delta", False) and not state["completed"]:
                    state["finish_reason"] = finish_reason
                    state["completed"] = True
                    cumulative_val = delta.get("cumulative_logprob")
                    if cumulative_val is not None:
                        try:
                            state["cumulative_logprob"] = float(cumulative_val)
                        except (TypeError, ValueError):
                            logger.debug(
                                "Failed to parse cumulative_logprob for request %d prompt %d candidate %d: %s",
                                request_id,
                                prompt_index,
                                candidate_index,
                                cumulative_val,
                            )
                    generation_len_val = delta.get("generation_len")
                    if generation_len_val is not None:
                        try:
                            state["generation_len"] = int(generation_len_val)
                        except (TypeError, ValueError):
                            logger.debug(
                                "Failed to parse generation_len for request %d prompt %d candidate %d: %s",
                                request_id,
                                prompt_index,
                                candidate_index,
                                generation_len_val,
                            )
                    remaining_sequences -= 1
                    logger.debug(
                        "Sequence %s for request %d (prompt %d, candidate %d) completed with reason %s",
                        sequence_id,
                        request_id,
                        prompt_index,
                        candidate_index,
                        finish_reason,
                    )
            finally:
                queue.task_done()
                release_delta_resources(delta)

    total_prompt_tokens = sum(prompt_token_totals)

    total_completion_tokens = sum(
        candidate_state["completion_tokens"]
        for prompt_state in prompt_states
        for candidate_state in prompt_state
    )

    selections: list[list[tuple[float, int, dict[str, Any]]]] = []
    for prompt_idx, candidate_states in enumerate(prompt_states):
        if not candidate_states:
            selections.append([])
            continue
        fanout = (
            prompt_fanout_counts[prompt_idx]
            if prompt_idx < len(prompt_fanout_counts)
            else len(candidate_states)
        )
        final_target = (
            prompt_final_counts[prompt_idx]
            if prompt_idx < len(prompt_final_counts)
            else len(candidate_states)
        )
        if final_target <= 0:
            final_target = 1
        final_target = min(final_target, len(candidate_states))

        scored_candidates: list[tuple[float, int, dict[str, Any]]] = []
        for idx, candidate_state in enumerate(candidate_states):
            cumulative = candidate_state.get("cumulative_logprob")
            generation_len = candidate_state.get("generation_len", 0)
            if cumulative is not None and generation_len and generation_len > 0:
                score = float(cumulative) / float(generation_len)
            else:
                score = float("-inf")
            candidate_state["average_logprob"] = score
            scored_candidates.append((score, idx, candidate_state))

        if final_target < fanout:
            scored_candidates.sort(key=lambda entry: (-entry[0], entry[1]))
            selections.append(scored_candidates[:final_target])
        else:
            selections.append(scored_candidates)

    choices: list[ChatCompletionChoice] = []
    single_prompt = len(prompt_states) == 1
    running_index = 0
    for _, selected_entries in enumerate(selections):
        for rank_within_prompt, (_, _, candidate_state) in enumerate(selected_entries):
            logprobs_object = None
            if candidate_state["logprobs_entries"]:
                logprobs_object = ChatCompletionLogProbs.LogProbsContent(
                    content=candidate_state["logprobs_entries"]
                )

            choice_content = (
                candidate_state["error_message"]
                if candidate_state["error_message"] is not None
                else candidate_state["content"]
            )

            choice_index = rank_within_prompt if single_prompt else running_index
            running_index += 1

            choices.append(
                ChatCompletionChoice(
                    index=choice_index,
                    message=ChatMessage(role="assistant", content=choice_content),
                    finish_reason=candidate_state["finish_reason"],
                    logprobs=logprobs_object,
                )
            )

    return {
        "choices": choices,
        "prompt_tokens": total_prompt_tokens,
        "completion_tokens": total_completion_tokens,
    }


async def stream_response_generator(
    request_id: int,
    queue: asyncio.Queue[ResponseDeltaDict],
    ipc_state: IPCStateDep,
    model_name: str,
    expected_sequences: int,
) -> AsyncIterable[dict[str, str]]:
    """
    Generates Server-Sent Events (SSE) from the response queue.

    This generator handles streaming responses from the C++ engine, converting
    them to OpenAI-compatible SSE format. It ensures proper cleanup of resources
    even in case of errors or client disconnections.
    """
    completion_tokens_by_candidate: defaultdict[tuple[int, int], int] = defaultdict(int)
    completed_sequences: set[int] = set()
    completed_candidate_slots: set[tuple[int, int]] = set()
    remaining_sequences = expected_sequences
    chat_completion_id = generate_chat_completion_id()
    created_at = get_current_timestamp()
    logger.info("Starting SSE stream for request %d", request_id)

    while True:
        try:
            delta_dict = await asyncio.wait_for(queue.get(), timeout=30.0)
        except TimeoutError:
            logger.error(
                "Timeout waiting for delta for streaming request %d", request_id
            )
            break
        else:
            try:
                prompt_index = int(delta_dict.get("prompt_index", 0) or 0)
                candidate_index = int(delta_dict.get("candidate_index", 0) or 0)
                candidate_key = (prompt_index, candidate_index)

                normalise_delta_payload(delta_dict)

                chunk_choice = ChatCompletionChunkChoice(
                    index=candidate_index,
                    delta=ChatMessage(
                        # only send assistant role for the first chunk
                        role="assistant"
                        if completion_tokens_by_candidate[candidate_key] <= 0
                        else None,
                        content=delta_dict.get("content", None),
                    ),
                    finish_reason=None,
                    logprobs=None,
                )
                chunk = ChatCompletionChunk(
                    id=chat_completion_id,
                    created=created_at,
                    model=model_name,
                    choices=[chunk_choice],
                )

                token_list = delta_dict.get("tokens", [])
                completion_tokens_by_candidate[candidate_key] += len(token_list)

                yield {"data": chunk.model_dump_json(exclude_none=True)}

                if delta_dict.get("is_final_delta", False):
                    finish_reason = delta_dict.get("finish_reason", "stop")
                    sequence_id = delta_dict.get("sequence_id")

                    if (
                        sequence_id is not None
                        and sequence_id not in completed_sequences
                    ):
                        completed_sequences.add(sequence_id)
                        remaining_sequences -= 1
                    elif (
                        sequence_id is None
                        and candidate_key not in completed_candidate_slots
                    ):
                        completed_candidate_slots.add(candidate_key)
                        remaining_sequences -= 1

                    final_chunk_choice = ChatCompletionChunkChoice(
                        index=candidate_index,
                        delta=ChatMessage(role=None, content=None),
                        finish_reason=finish_reason,
                        logprobs=None,
                    )
                    final_chunk = ChatCompletionChunk(
                        id=chat_completion_id,
                        created=created_at,
                        model=model_name,
                        choices=[final_chunk_choice],
                    )
                    yield {"data": final_chunk.model_dump_json(exclude_none=True)}
                    if remaining_sequences <= 0:
                        break
            finally:
                queue.task_done()
                release_delta_resources(delta_dict)

    yield {"data": "[DONE]"}
    logger.info("SSE stream for request %d completed", request_id)



import asyncio
import logging
import random
from typing import Any

from fastapi import APIRouter, HTTPException, status
from fastapi.responses import JSONResponse

from orchard.app.ipc_dispatch import QueueRegistration
from orchard.ipc.serialization import _build_request_payload
from orchard.ipc.utils import (
    ResponseDeltaDict,
    release_delta_resources,
)
from orchard.server.dependencies import IPCStateDep, ModelRegistryDep
from orchard.server.exceptions import InferenceError
from orchard.server.models.chat.logprobs import ChatCompletionLogProbs
from orchard.server.models.chat.output import ChatCompletionChoice
from orchard.server.models.completions import (
    CompletionChoice,
    CompletionRequest,
    CompletionResponse,
    CompletionUsage,
)
from orchard.server.routes._common import _ModelNotReadyError, resolve_model
from orchard.server.routes.chat import (
    gather_non_streaming_batch_response,
)

logger = logging.getLogger(__name__)

completions_router = APIRouter()


@completions_router.post(
    "/completions",
    response_model=CompletionResponse,
    summary="Create a text completion",
    tags=["Completions"],
)
async def handle_completion_request(
    request: CompletionRequest,
    ipc_state: IPCStateDep,
    model_registry: ModelRegistryDep,
) -> CompletionResponse | JSONResponse:
    """Handles requests to the `/v1/completions` endpoint."""
    logger.info("Handling completion request for model: %s", request.model)

    if request.stream:
        logger.warning("Streaming requested for /completions but not supported.")
        raise HTTPException(
            status_code=status.HTTP_501_NOT_IMPLEMENTED,
            detail="Streaming is not supported in this version.",
        )

    prompts = _normalize_prompt_inputs(request.prompt)

    try:
        canonical_id, model_info = await resolve_model(model_registry, request.model)
    except _ModelNotReadyError as exc:
        return exc.response

    formatter = model_info.formatter
    model_path = model_info.model_path

    current_request_id = await ipc_state.get_next_request_id()
    response_channel_id = ipc_state.response_channel_id or current_request_id
    response_queue: asyncio.Queue[ResponseDeltaDict] = asyncio.Queue()
    loop = asyncio.get_running_loop()
    ipc_state.active_request_queues[current_request_id] = QueueRegistration(
        loop=loop,
        queue=response_queue,
    )

    fanout_counts: list[int] = []
    final_candidate_counts: list[int] = []
    prompt_payloads: list[dict[str, Any]] = []

    for prompt_text in prompts:
        effective_prompt = prompt_text
        if request.apply_chat_template:
            try:
                effective_prompt = formatter.apply_template(
                    [{"role": "user", "content": prompt_text}]
                )
            except Exception as exc:  # pragma: no cover - template errors
                logger.exception("Failed to apply chat template: %s", exc)
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail="Failed to render prompt template.",
                ) from exc

        prompt_bytes = effective_prompt.encode("utf-8")
        layout_segments = [
            {
                "type": "text",
                "length": len(prompt_bytes),
            }
        ]

        rng_seed = random.randint(0, 2**32 - 1)
        normalized_top_k = request.top_k if request.top_k > 0 else -1
        num_candidates = request.best_of if request.best_of is not None else request.n
        num_candidates = max(1, num_candidates)
        final_candidates = max(1, request.n)
        if final_candidates > num_candidates:
            final_candidates = num_candidates

        prompt_payloads.append(
            {
                "prompt_bytes": prompt_bytes,
                "image_buffers": [],
                "layout": layout_segments,
                "sampling_params": {
                    "temperature": request.temperature,
                    "top_p": request.top_p,
                    "top_k": normalized_top_k,
                    "min_p": request.min_p,
                    "rng_seed": rng_seed,
                },
                "logits_params": {
                    "top_logprobs": request.logprobs or 0,
                    "frequency_penalty": 0.0,
                    "logit_bias": {},
                    "presence_penalty": 0.0,
                    "repetition_context_size": 60,
                    "repetition_penalty": 1.0,
                },
                "max_generated_tokens": request.max_completion_tokens,
                "stop_sequences": [],
                "tool_schemas_json": "",
                "response_format_json": "",
                "num_candidates": num_candidates,
                "best_of": num_candidates,
                "final_candidates": final_candidates,
                "tool_calling_tokens": formatter.get_tool_calling_tokens(),
            }
        )
        fanout_counts.append(num_candidates)
        final_candidate_counts.append(final_candidates)

    if not prompt_payloads:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="At least one prompt is required.",
        )

    try:
        request_bytes = _build_request_payload(
            request_id=current_request_id,
            model_id=canonical_id,
            model_path=model_path,
            request_type="generation",
            response_channel_id=response_channel_id,
            prompts=prompt_payloads,
        )

        socket = ipc_state.request_socket
        if socket is None:
            raise RuntimeError("Request socket is not initialized.")

        await socket.asend(request_bytes)
        logger.info(
            "Submitted completions request %d with %d prompts.",
            current_request_id,
            len(prompt_payloads),
        )

        response_data = await gather_non_streaming_batch_response(
            current_request_id,
            response_queue,
            fanout_counts,
            final_candidate_counts,
        )

        completion_choices = _convert_chat_choices(response_data["choices"])
        usage = CompletionUsage(
            input_tokens=response_data["prompt_tokens"],
            output_tokens=response_data["completion_tokens"],
            total_tokens=response_data["prompt_tokens"]
            + response_data["completion_tokens"],
        )
        response = CompletionResponse(
            model=request.model,
            choices=completion_choices,
            usage=usage,
        )
        logger.info("Completion request %d completed successfully.", current_request_id)
        return response
    except HTTPException:
        raise
    except InferenceError as exc:
        logger.error(
            "Inference error processing completion request %d: %s",
            current_request_id,
            exc,
        )
        raise HTTPException(
            status_code=status.HTTP_502_BAD_GATEWAY,
            detail=str(exc),
        ) from exc
    except Exception as exc:  # pragma: no cover - defensive
        logger.exception(
            "Unexpected error while processing completion request %d: %s",
            current_request_id,
            exc,
        )
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="An unexpected error occurred during completion.",
        ) from exc
    finally:
        binding = ipc_state.active_request_queues.pop(current_request_id, None)
        if binding is not None:
            queue = binding.queue
            try:
                while True:
                    leftover = queue.get_nowait()
                    release_delta_resources(leftover)
                    queue.task_done()
            except asyncio.QueueEmpty:
                pass


def _normalize_prompt_inputs(prompt: str | list[str]) -> list[str]:
    if isinstance(prompt, str):
        return [prompt]
    if isinstance(prompt, list) and prompt:
        if not all(isinstance(entry, str) for entry in prompt):
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="All prompt entries must be strings.",
            )
        return prompt
    raise HTTPException(
        status_code=status.HTTP_400_BAD_REQUEST,
        detail="Invalid prompt format. Expecting string or non-empty list of strings.",
    )


def _convert_chat_choices(
    chat_choices: list[ChatCompletionChoice],
) -> list[CompletionChoice]:
    completion_choices: list[CompletionChoice] = []
    for choice in chat_choices:
        message_content = ""
        if choice.message and choice.message.content is not None:
            message_content = str(choice.message.content)
        completion_choices.append(
            CompletionChoice(
                index=choice.index,
                text=message_content,
                logprobs=_extract_logprob_values(choice.logprobs),
                finish_reason=choice.finish_reason,
            )
        )
    return completion_choices


def _extract_logprob_values(
    logprob_content: ChatCompletionLogProbs.LogProbsContent | None,
) -> list[float] | None:
    if not logprob_content or not logprob_content.content:
        return None
    values = []
    for entry in logprob_content.content:
        try:
            values.append(float(entry.logprob))
        except (TypeError, ValueError):
            continue
    return values or None



import asyncio
import json
import logging
import time
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING

from huggingface_hub import snapshot_download

from orchard.app.model_resolver import (
    ModelResolutionError,
    ModelResolver,
    ResolvedModel,
)
from orchard.engine.multiprocess import pid_is_alive, read_pid_file
from orchard.formatter import ChatFormatter

if TYPE_CHECKING:
    from orchard.app.ipc_dispatch import IPCState

logger = logging.getLogger(__name__)


class ModelLoadState(Enum):
    IDLE = auto()
    DOWNLOADING = auto()
    ACTIVATING = auto()
    LOADING = auto()
    READY = auto()
    FAILED = auto()


@dataclass(slots=True)
class ModelInfo:
    model_id: str
    model_path: str
    formatter: ChatFormatter
    capabilities: dict[str, list[int]] | None = None


@dataclass(slots=True)
class ModelEntry:
    state: ModelLoadState = ModelLoadState.IDLE
    info: ModelInfo | None = None
    error: str | None = None
    event: asyncio.Event = field(default_factory=asyncio.Event)
    task: asyncio.Task[None] | None = None
    resolved: ResolvedModel | None = None
    bytes_downloaded: int | None = None
    bytes_total: int | None = None
    activation_future: asyncio.Future | None = None
    activation_loop: asyncio.AbstractEventLoop | None = None


class ModelRegistry:
    def __init__(self, ipc_state: "IPCState") -> None:
        self._entries: dict[str, ModelEntry] = {}
        self._lock = asyncio.Lock()
        self._resolver = ModelResolver()
        self._alias_cache: dict[str, str] = {}
        self._ipc_state = ipc_state

    async def ensure_ready(
        self, requested_model_id: str, *, timeout: float | None = None
    ) -> ModelInfo:
        """Ensure a model is loaded and return the ModelInfo."""
        return await self.ensure_loaded(requested_model_id, timeout=timeout)

    def ensure_ready_sync(
        self, requested_model_id: str, *, timeout: float | None = None
    ) -> ModelInfo:
        """Blocking wrapper around ensure_ready for synchronous contexts."""

        return asyncio.run(self.ensure_ready(requested_model_id, timeout=timeout))

    async def ensure_loaded(
        self, requested_model_id: str, timeout: float | None = None
    ) -> ModelInfo:
        """Ensure a model is staged locally and activated on the engine."""
        state, canonical_id = await self.schedule_model(requested_model_id)
        # Wait for local readiness (download + formatter)
        state, info, error = await self.await_model(canonical_id, timeout)
        if state == ModelLoadState.FAILED or not info:
            detail = error or f"Model '{canonical_id}' failed to load."
            raise RuntimeError(detail)
        if state == ModelLoadState.READY:
            return info

        # Activation phase (IPC load command)
        async with self._lock:
            entry = self._entries[canonical_id]
            if entry.state == ModelLoadState.READY and entry.info:
                return entry.info

            loop = asyncio.get_running_loop()
            should_send_command = True
            # Reuse an in-flight activation if one exists
            if entry.state == ModelLoadState.ACTIVATING and entry.activation_future:
                waiter = entry.activation_future
                should_send_command = False
            else:
                # Cancel any stale waiter
                if entry.activation_future and not entry.activation_future.done():
                    entry.activation_future.cancel()
                entry.activation_loop = loop
                entry.activation_future = loop.create_future()
                entry.state = ModelLoadState.ACTIVATING
                waiter = entry.activation_future

        if should_send_command:
            await self._send_load_model_command(
                requested_id=requested_model_id,
                canonical_id=canonical_id,
                info=info,
            )

        try:
            await self._await_activation_with_engine_liveness(
                canonical_id=canonical_id,
                waiter=waiter,
                timeout=timeout,
            )
        except Exception as exc:
            async with self._lock:
                entry = self._entries.get(canonical_id)
                if entry and entry.state != ModelLoadState.READY:
                    entry.state = ModelLoadState.FAILED
                    entry.error = str(exc)
                    entry.activation_future = None
                    entry.activation_loop = None
            raise

        ready_info = self.get_if_ready(canonical_id)
        if not ready_info:
            raise RuntimeError(f"Model '{canonical_id}' failed to activate.")
        return ready_info

    async def schedule_model(
        self, requested_model_id: str, *, force_reload: bool = False
    ) -> tuple[ModelLoadState, str]:
        """Ensure a model is loading or ready and return the canonical identifier."""

        resolved = self._resolver.resolve(requested_model_id)
        canonical_id = resolved.canonical_id
        self._alias_cache[requested_model_id.lower()] = canonical_id
        self._alias_cache.setdefault(canonical_id.lower(), canonical_id)

        async with self._lock:
            entry = self._entries.get(canonical_id)
            if entry is None:
                entry = ModelEntry()
                self._entries[canonical_id] = entry

            if entry.state == ModelLoadState.READY and not force_reload:
                return ModelLoadState.READY, canonical_id

            if (
                entry.state
                in (
                    ModelLoadState.LOADING,
                    ModelLoadState.DOWNLOADING,
                    ModelLoadState.ACTIVATING,
                )
                and not force_reload
            ):
                return entry.state, canonical_id

            if entry.state == ModelLoadState.FAILED and not force_reload:
                return ModelLoadState.FAILED, canonical_id

            entry.error = None
            entry.info = None
            entry.event = asyncio.Event()
            entry.resolved = resolved
            entry.bytes_downloaded = None
            entry.bytes_total = None
            entry.activation_future = None
            entry.activation_loop = None

            # If the model is already local (or in HF cache), build formatter immediately.
            if resolved.source in {"local", "hf_cache"} or (
                resolved.model_path and (resolved.model_path / "config.json").exists()
            ):
                try:
                    formatter = await asyncio.to_thread(
                        ChatFormatter, str(resolved.model_path)
                    )
                    info = ModelInfo(
                        model_id=resolved.canonical_id,
                        model_path=str(resolved.model_path),
                        formatter=formatter,
                    )
                    entry.info = info
                    entry.state = ModelLoadState.LOADING
                    entry.event.set()
                    return entry.state, canonical_id
                except Exception as exc:
                    entry.error = str(exc)
                    entry.state = ModelLoadState.FAILED
                    entry.event.set()
                    return ModelLoadState.FAILED, canonical_id

            # Otherwise, schedule async download + formatter build.
            entry.state = ModelLoadState.DOWNLOADING

            async def loader() -> None:
                loop = asyncio.get_running_loop()

                async def _update_progress(
                    bytes_downloaded: int, bytes_total: int
                ) -> None:
                    async with self._lock:
                        entry.bytes_downloaded = bytes_downloaded
                        entry.bytes_total = bytes_total

                def _progress_cb(bytes_downloaded: int, bytes_total: int) -> None:
                    loop.call_soon_threadsafe(
                        asyncio.create_task,
                        _update_progress(bytes_downloaded, bytes_total),
                    )

                try:
                    download_path = await asyncio.to_thread(
                        snapshot_download,
                        resolved.hf_repo or resolved.canonical_id,
                        local_files_only=False,
                        progress_callback=_progress_cb,  # type: ignore [reportCallIssue]
                    )
                    await _update_progress(
                        entry.bytes_total or 0, entry.bytes_total or 0
                    )
                    formatter = await asyncio.to_thread(
                        ChatFormatter, str(download_path)
                    )
                    info = ModelInfo(
                        model_id=resolved.canonical_id,
                        model_path=str(download_path),
                        formatter=formatter,
                    )
                    async with self._lock:
                        entry.info = info
                        entry.state = ModelLoadState.LOADING
                        entry.resolved = ResolvedModel(
                            canonical_id=resolved.canonical_id,
                            model_path=Path(download_path),
                            source="hf_cache",
                            metadata=resolved.metadata,
                            hf_repo=resolved.hf_repo,
                        )
                except (
                    asyncio.CancelledError
                ):  # pragma: no cover - cooperative cancellation
                    raise
                except Exception as exc:  # pragma: no cover - best effort logging
                    async with self._lock:
                        entry.error = str(exc)
                        entry.state = ModelLoadState.FAILED
                        entry.info = None
                finally:
                    entry.event.set()

            entry.task = asyncio.create_task(loader())
            return ModelLoadState.DOWNLOADING, canonical_id

    async def await_model(
        self, model_id: str, timeout: float | None = None
    ) -> tuple[ModelLoadState, ModelInfo | None, str | None]:
        canonical_id = self._canonicalize(model_id)
        if canonical_id is None:
            raise ValueError(f"Model '{model_id}' has not been scheduled")

        async with self._lock:
            entry = self._entries.get(canonical_id)
            if entry is None:
                raise ValueError(f"Model '{canonical_id}' has not been scheduled")
            event = entry.event

        try:
            if timeout is None:
                await event.wait()
            else:
                await asyncio.wait_for(event.wait(), timeout)
        except TimeoutError:
            return ModelLoadState.LOADING, None, None

        async with self._lock:
            entry = self._entries[canonical_id]
            return entry.state, entry.info, entry.error

    def get_if_ready(self, model_id: str) -> ModelInfo | None:
        canonical_id = self._canonicalize(model_id)
        if canonical_id is None:
            return None
        entry = self._entries.get(canonical_id)
        if entry and entry.state == ModelLoadState.READY:
            return entry.info
        return None

    def get_status(
        self, model_id: str
    ) -> tuple[ModelLoadState, str | None, dict | None]:
        canonical_id = self._canonicalize(model_id)
        if canonical_id is None:
            return ModelLoadState.IDLE, None, None
        entry = self._entries.get(canonical_id)
        if entry is None:
            return ModelLoadState.IDLE, None, None
        progress = None
        if entry.bytes_downloaded is not None or entry.bytes_total is not None:
            progress = {
                "bytes_downloaded": entry.bytes_downloaded,
                "bytes_total": entry.bytes_total,
            }
        return entry.state, entry.error, progress

    def get_error(self, model_id: str) -> str | None:
        canonical_id = self._canonicalize(model_id)
        if canonical_id is None:
            return None
        entry = self._entries.get(canonical_id)
        if entry and entry.state == ModelLoadState.FAILED:
            return entry.error
        return None

    def list_models(self) -> list[dict[str, str]]:
        """List all currently loaded models."""
        catalog = []
        for canonical_id, entry in self._entries.items():
            if entry.resolved is None:
                continue
            payload = dict(entry.resolved.metadata)
            payload["canonical_id"] = canonical_id
            payload["model_path"] = str(entry.resolved.model_path)
            payload["source"] = entry.resolved.source
            payload["hf_repo"] = entry.resolved.hf_repo or ""
            payload["state"] = entry.state.name
            catalog.append(payload)
        return catalog

    def resolve(self, model_id: str) -> ResolvedModel:
        return self._resolver.resolve(model_id)

    async def get_info(self, model_id: str) -> ModelInfo:
        if (
            model_id in self._alias_cache
            and (canonical := self._alias_cache[model_id]) in self._entries
            and (info := self._entries[canonical].info) is not None
        ):
            return info

        return await self.ensure_ready(model_id)

    async def _send_load_model_command(
        self,
        *,
        requested_id: str,
        canonical_id: str,
        info: ModelInfo,
    ) -> None:
        """Issue the IPC load_model command and handle immediate responses."""
        ipc_state = self._ipc_state
        if not ipc_state or not ipc_state.management_socket:
            raise RuntimeError("IPC state is not initialized.")

        async with self._lock:
            entry = self._entries.get(canonical_id)
        if entry is None or entry.activation_future is None:
            await self._mark_activation_failed(
                canonical_id, "Activation waiter missing."
            )
            raise RuntimeError("Activation waiter missing.")

        command = {
            "type": "load_model",
            "requested_id": requested_id,
            "canonical_id": canonical_id,
            "model_path": info.model_path,
            "wait_for_completion": False,
        }
        payload = json.dumps(command).encode("utf-8")

        management_socket = ipc_state.management_socket
        assert management_socket is not None  # for type-checkers

        try:
            async with ipc_state.management_lock:
                await management_socket.asend(payload)
                reply_bytes = await management_socket.arecv()
        except Exception as exc:
            await self._mark_activation_failed(
                canonical_id, f"Failed to send load_model command: {exc}"
            )
            raise RuntimeError(
                f"Failed to send load_model command for '{requested_id}': {exc}"
            ) from exc

        try:
            response = json.loads(reply_bytes.decode("utf-8"))
        except Exception as exc:
            await self._mark_activation_failed(
                canonical_id, "Engine returned malformed management response."
            )
            raise RuntimeError(
                "Engine returned malformed management response."
            ) from exc

        status = response.get("status")
        if status == "ok":
            capabilities = self._parse_capabilities(response)
            try:
                self.update_capabilities(canonical_id, capabilities)
            finally:
                await self._complete_activation(canonical_id, capabilities is not None)
            return

        if status != "accepted":
            message = response.get("message", "unknown error")
            await self._mark_activation_failed(
                canonical_id, f"Engine rejected load_model: {message}"
            )
            raise RuntimeError(
                f"Engine rejected load_model for '{requested_id}': {message}"
            )

        # Accepted: wait for model_loaded event via handle_model_loaded
        async with self._lock:
            entry = self._entries[canonical_id]
            waiter = entry.activation_future

        if waiter is None:
            await self._mark_activation_failed(
                canonical_id, "Activation waiter missing."
            )
            raise RuntimeError("Activation waiter missing.")

    def update_capabilities(self, model_id: str, capabilities: dict | None) -> None:
        """Update persisted capabilities for a given model."""
        if not capabilities:
            return

        canonical_id = self._canonicalize(model_id) or model_id
        entry = self._entries.get(canonical_id)
        if not entry or entry.info is None:
            logger.warning("Received capabilities for unknown model_id '%s'.", model_id)
            return

        normalized: dict[str, list[int]] = {}
        for name, value in capabilities.items():
            if isinstance(value, list | tuple):
                normalized[str(name)] = [int(v) for v in value]
            else:
                normalized[str(name)] = [int(value)]

        entry.info.capabilities = normalized

    def handle_model_loaded(self, payload: dict) -> None:
        """Handle model_loaded events emitted by the engine."""
        model_id = payload.get("model_id")
        if not model_id:
            logger.warning("Received model_loaded event without model_id.")
            return

        capabilities = payload.get("capabilities")
        try:
            self.update_capabilities(model_id, capabilities)
        except Exception:
            logger.exception("Failed to update capabilities for model '%s'.", model_id)

        entry = self._entries.get(model_id)
        if not entry:
            logger.debug("Received model_loaded for unknown model '%s'.", model_id)
            return

        waiter = entry.activation_future
        loop = entry.activation_loop
        if entry.state != ModelLoadState.ACTIVATING:
            return

        entry.state = ModelLoadState.READY
        if loop and waiter and not waiter.done():
            loop.call_soon_threadsafe(waiter.set_result, payload)
        entry.activation_future = None
        entry.activation_loop = None

    def handle_model_load_failed(self, payload: dict) -> None:
        """Handle model_load_failed events emitted by the engine."""
        model_id = payload.get("model_id")
        if not model_id:
            logger.warning("Received model_load_failed event without model_id.")
            return

        error = payload.get("error")
        if not isinstance(error, str) or not error:
            error = f"Model '{model_id}' failed to load."

        canonical_id = self._canonicalize(model_id) or model_id
        entry = self._entries.get(canonical_id)
        if not entry:
            logger.debug(
                "Received model_load_failed for unknown model '%s'.", model_id
            )
            return

        if entry.state != ModelLoadState.ACTIVATING:
            return

        self._set_activation_failed(entry, error)

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    @staticmethod
    def _engine_pid_file() -> Path:
        home = Path.home()
        mac_cache = home / "Library" / "Caches"
        base = mac_cache if mac_cache.exists() else home / ".cache"
        return base / "com.theproxycompany" / "engine.pid"

    @classmethod
    def _engine_pid_alive(cls) -> tuple[bool | None, int | None]:
        pid_file = cls._engine_pid_file()
        pid = read_pid_file(pid_file)
        if pid is None:
            return None, None
        return pid_is_alive(pid), pid

    async def _await_activation_with_engine_liveness(
        self,
        *,
        canonical_id: str,
        waiter: asyncio.Future,
        timeout: float | None,
    ) -> None:
        poll_interval_s = 0.25
        deadline = None if timeout is None else (time.monotonic() + timeout)

        while True:
            if waiter.done():
                await waiter
                return

            alive, pid = self._engine_pid_alive()
            if alive is False:
                pid_suffix = f" (pid={pid})" if pid else ""
                raise RuntimeError(
                    f"Engine process exited while activating '{canonical_id}'{pid_suffix}."
                )

            if deadline is not None:
                remaining = deadline - time.monotonic()
                if remaining <= 0:
                    raise TimeoutError(
                        f"Timed out waiting for model '{canonical_id}' activation."
                    )
                step_timeout = min(poll_interval_s, remaining)
            else:
                step_timeout = poll_interval_s

            try:
                await asyncio.wait_for(asyncio.shield(waiter), timeout=step_timeout)
                await waiter
                return
            except TimeoutError:
                continue

    def _parse_capabilities(self, response: dict) -> dict | None:
        try:
            data_field = response.get("data", {})
            if isinstance(data_field, dict):
                load_model_data = data_field.get("load_model") or {}
                if isinstance(load_model_data, dict):
                    return load_model_data.get("capabilities")
        except Exception:
            logger.exception("Failed to parse capabilities from management response.")
        return None

    async def _complete_activation(
        self, model_id: str, capabilities_updated: bool
    ) -> None:
        async with self._lock:
            entry = self._entries.get(model_id)
            if not entry:
                return
            entry.state = ModelLoadState.READY
            if (
                entry.activation_future
                and not entry.activation_future.done()
                and entry.activation_loop
            ):
                entry.activation_loop.call_soon_threadsafe(
                    entry.activation_future.set_result,
                    {"capabilities": capabilities_updated},
                )
            entry.activation_future = None
            entry.activation_loop = None

    async def _mark_activation_failed(self, model_id: str, reason: str) -> None:
        async with self._lock:
            entry = self._entries.get(model_id)
            if not entry:
                return
            self._set_activation_failed(entry, reason)

    @staticmethod
    def _set_activation_failed(entry: ModelEntry, reason: str) -> None:
        entry.error = reason
        entry.state = ModelLoadState.FAILED
        if entry.activation_future and not entry.activation_future.done():
            if entry.activation_loop:
                entry.activation_loop.call_soon_threadsafe(
                    entry.activation_future.set_exception, RuntimeError(reason)
                )
            else:
                entry.activation_future.set_exception(RuntimeError(reason))
        entry.activation_future = None
        entry.activation_loop = None

    def _canonicalize(self, model_id: str) -> str | None:
        """Convert a model ID to its canonical form if known."""
        if model_id in self._entries:
            return model_id
        lower = model_id.lower()
        if lower in self._alias_cache:
            return self._alias_cache[lower]
        return None


__all__ = [
    "ModelLoadState",
    "ModelRegistry",
    "ModelResolutionError",
]



import json

import httpx
import pytest

pytestmark = pytest.mark.asyncio

MODEL_ID = "moondream3"


@pytest.mark.parametrize("temperature", [0.0])
async def test_chat_completion_structured_json_response(live_server, temperature):
    server_url = live_server
    schema = {
        "type": "object",
        "properties": {
            "color": {
                "type": "object",
                "properties": {
                    "R": {"type": "integer", "minimum": 0, "maximum": 255},
                    "G": {"type": "integer", "minimum": 0, "maximum": 255},
                    "B": {"type": "integer", "minimum": 0, "maximum": 255},
                },
                "required": ["R", "G", "B"],
            },
            "confidence": {"type": "number", "minimum": 0.0, "maximum": 1.0},
        },
        "required": ["color", "confidence"],
    }
    payload = {
        "model": MODEL_ID,
        "messages": [
            {
                "role": "user",
                "content": f"Respond with a JSON object with a rgb(r, g, b) color and a confidence score. use this schema: {json.dumps(schema, indent=2)}. Make it pretty, this should reflect your internal associations with the color.",
            }
        ],
        "temperature": temperature,
        "stream": False,
        "logprobs": True,
        "top_logprobs": 3,
        "response_format": {
            "type": "json_schema",
            "json_schema": {
                "name": "color_summary",
                "strict": True,
                "schema": schema,
            },
        },
    }

    async with httpx.AsyncClient(timeout=60.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions",
            json=payload,
        )

    assert response.status_code == 200
    body = response.json()
    assert body.get("choices")

    content = body["choices"][0]["message"]["content"]
    assert content, "Expected structured content in completion."

    start = content.find("{")
    end = content.rfind("}")
    if start == -1 or end == -1 or end <= start:
        raise ValueError(f"Invalid JSON in completion: {content}")

    json_payload = content[start : end + 1]
    parsed = json.loads(json_payload)
    assert isinstance(parsed, dict)
    assert "color" in parsed
    assert isinstance(parsed["color"], dict)
    assert "R" in parsed["color"]
    assert isinstance(parsed["color"]["R"], int | float)
    assert "G" in parsed["color"]
    assert isinstance(parsed["color"]["G"], int | float)
    assert "B" in parsed["color"]
    assert isinstance(parsed["color"]["B"], int | float)
    if "confidence" in parsed:
        assert isinstance(parsed["confidence"], int | float)
    r = int(parsed["color"]["R"])
    g = int(parsed["color"]["G"])
    b = int(parsed["color"]["B"])
    if "confidence" in parsed:
        confidence = parsed["confidence"]
        opacity = int(confidence * 255)
        colored_text = f"\033[38;2;{r};{g};{b};{opacity}m{MODEL_ID}'s color is rgb({r}, {g}, {b}) with confidence {confidence} at temperature {temperature}.\033[0m"
    else:
        colored_text = f"\033[38;2;{r};{g};{b}m{MODEL_ID}'s color is rgb({r}, {g}, {b}) at temperature {temperature}. No confidence score provided.\033[0m"
    print(colored_text)



import httpx
import pytest

pytestmark = pytest.mark.asyncio

MODEL_ID = "meta-llama/Llama-3.1-8B-Instruct"


async def test_chat_completion_first_token(live_server):
    """
    Tests a basic, non-streaming chat completion request to the live server.
    Verifies that the system can process a request and return a valid response.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [{"role": "user", "content": "Hello!"}],
        "max_completion_tokens": 1,  # We only care about the first token for now
        "temperature": 1.0,  # Use greedy sampling which is implemented
        "stream": False,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions", json=request_payload
        )

    # --- Assertions ---
    assert response.status_code == 200
    response_data = response.json()

    assert "id" in response_data
    assert "choices" in response_data
    assert isinstance(response_data["choices"], list)
    assert len(response_data["choices"]) > 0

    choice = response_data["choices"][0]
    assert "message" in choice
    assert "content" in choice["message"]
    # We fixed the bug, so content should not be empty.
    assert choice["message"]["content"] is not None
    assert len(choice["message"]["content"]) > 0

    assert "finish_reason" in choice
    # For a 1-token generation, the reason should be 'length'.
    assert choice["finish_reason"].lower() in ["length", "stop"]


async def test_chat_completion_multi_token(live_server):
    """
    Tests a non-streaming request that requires multiple tokens to be generated.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [
            {
                "role": "user",
                "content": "Provide one friendly sentence introducing yourself.",
            }
        ],
        "max_completion_tokens": 64,
        "temperature": 0.0,  # Greedy for deterministic output
        "stream": False,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions", json=request_payload
        )

    assert response.status_code == 200
    response_data = response.json()

    assert len(response_data["choices"]) > 0
    # contents should be deterministic
    content = response_data["choices"][0]["message"]["content"]
    print(content)
    assert content is not None
    assert len(content) > 0



from __future__ import annotations

from dataclasses import dataclass

from orchard.server.models.chat.format import (
    ChatCompletionJsonObjectResponseFormat,
    ChatCompletionJSONSchemaResponseFormat,
    ChatCompletionTextResponseFormat,
)
from orchard.server.models.chat.request import (
    ChatCompletionRequest,
    ChatMessage,
)
from orchard.server.models.chat.tools import ChatCompletionTool
from orchard.server.models.reasoning import ReasoningEffort


@dataclass
class NormalizedChatInstance:
    prompt_index: int
    messages: list[ChatMessage]
    max_completion_tokens: int | None
    temperature: float
    top_p: float | None
    top_k: int | None
    min_p: float | None
    logprobs: bool
    top_logprobs: int | None
    tools: list[ChatCompletionTool] | None
    response_format: (
        ChatCompletionTextResponseFormat
        | ChatCompletionJSONSchemaResponseFormat
        | ChatCompletionJsonObjectResponseFormat
        | None
    )
    stop_sequences: list[str]
    best_of: int
    final_candidates: int
    task: str | None = None
    reasoning_effort: ReasoningEffort | None = None


def normalize_chat_request(
    request: ChatCompletionRequest,
) -> list[NormalizedChatInstance]:
    batch_size = request.batch_size

    max_tokens = request.get_normalized_field("max_completion_tokens")
    temperatures = request.get_normalized_field("temperature")
    top_ps = request.get_normalized_field("top_p")
    top_ks = request.get_normalized_field("top_k")
    min_ps = request.get_normalized_field("min_p")
    logprobs_flags = request.get_normalized_field("logprobs")
    top_logprobs = request.get_normalized_field("top_logprobs")
    tools_list = request.get_normalized_field("tools")
    response_formats = request.get_normalized_field("response_format")
    stop_sequences = request.get_normalized_field("stop")
    task_values = request.get_normalized_field("task")
    candidate_counts = request.get_normalized_field("n")
    best_of_counts = request.get_normalized_field("best_of")
    reasoning_efforts = request.get_normalized_field("reasoning_effort")

    instances: list[NormalizedChatInstance] = []
    for idx in range(batch_size):
        temp_value = temperatures[idx] if temperatures[idx] is not None else 1.0
        top_p_value = top_ps[idx] if top_ps[idx] is not None else 1.0
        top_k_value = top_ks[idx]
        min_p_value = min_ps[idx] if min_ps[idx] is not None else 0.0
        logprobs_enabled = (
            bool(logprobs_flags[idx]) if logprobs_flags[idx] is not None else False
        )
        top_logprobs_value = top_logprobs[idx] if logprobs_enabled else 0
        tools_value = tools_list[idx] if tools_list[idx] else None
        response_format_value = response_formats[idx]
        stops_value = stop_sequences[idx] if stop_sequences[idx] else []
        candidate_count = (
            int(candidate_counts[idx]) if candidate_counts[idx] is not None else 1
        )
        if candidate_count < 1:
            raise ValueError("n must be at least 1 for all batch entries.")
        best_of_value = (
            int(best_of_counts[idx])
            if best_of_counts[idx] is not None
            else candidate_count
        )
        messages = request.messages[idx]
        if not isinstance(messages, list):
            messages = [messages]

        instances.append(
            NormalizedChatInstance(
                prompt_index=idx,
                messages=messages,
                max_completion_tokens=max_tokens[idx],
                temperature=float(temp_value),
                top_p=float(top_p_value) if top_p_value is not None else None,
                top_k=int(top_k_value) if top_k_value is not None else None,
                min_p=float(min_p_value) if min_p_value is not None else None,
                logprobs=logprobs_enabled,
                top_logprobs=int(top_logprobs_value) if top_logprobs_value else 0,
                tools=tools_value,
                response_format=response_format_value,
                stop_sequences=list(stops_value),
                best_of=best_of_value,
                final_candidates=candidate_count,
                task=task_values[idx],
                reasoning_effort=reasoning_efforts[idx],
            )
        )

    return instances


__all__ = ["NormalizedChatInstance", "normalize_chat_request"]



from typing import Annotated, Literal

from pydantic import BaseModel, Field, field_validator, model_validator

from orchard.server.models.reasoning import (
    DEFAULT_BOOLEAN_REASONING_EFFORT,
    ReasoningEffort,
    normalize_reasoning_effort,
)
from orchard.server.models.responses.format import ResponseFormat
from orchard.server.models.responses.tools import (
    Function,
    FunctionID,
    ToolUseMode,
)
from orchard.server.models.tools import ToolCall, ToolCallFunction

# --- Content Parts (for message content) ---


class InputText(BaseModel):
    """Text content for an input message."""

    type: Literal["input_text"]
    text: str = Field(description="Raw text content.")


class InputImage(BaseModel):
    """Image content for an input message."""

    type: Literal["input_image"]
    image_url: str = Field(
        description="Data URL containing Base64-encoded image bytes."
    )
    detail: Literal["low", "high", "auto"] = Field(
        default="auto",
        description="Image detail level for processing.",
    )


ContentPart = Annotated[InputText | InputImage, Field(discriminator="type")]


# --- Input Item Types (discriminated by `type` field) ---


class InputMessageItem(BaseModel):
    """A message item in the input."""

    type: Literal["message"] = "message"
    role: Literal["user", "assistant", "system", "developer", "tool"] = Field(
        description="Role of the message author."
    )
    content: str | list[ContentPart] = Field(
        description="Message content as raw text or structured content parts."
    )
    tool_calls: list[ToolCall] | None = Field(
        default=None,
        description="Tool calls made in this message (assistant messages only).",
    )
    tool_call_id: str | None = Field(
        default=None,
        description="The call_id this tool message is responding to (tool messages only).",
    )


class InputFunctionCall(BaseModel):
    """A function call from a previous assistant turn."""

    type: Literal["function_call"] = "function_call"
    call_id: str = Field(description="Unique identifier for this function call.")
    name: str = Field(description="Name of the function that was called.")
    arguments: str = Field(description="JSON-encoded arguments passed to the function.")


class InputFunctionCallOutput(BaseModel):
    """The result of a function call."""

    type: Literal["function_call_output"] = "function_call_output"
    call_id: str = Field(
        description="The call_id of the function call this is responding to."
    )
    output: str = Field(description="The output of the function call.")


class InputReasoning(BaseModel):
    """Reasoning content from a previous turn."""

    type: Literal["reasoning"] = "reasoning"
    summary: list[dict] | None = Field(
        default=None,
        description="Summary of the reasoning (if available).",
    )
    encrypted_content: str | None = Field(
        default=None,
        description="Encrypted reasoning content for continuation.",
    )


InputItem = Annotated[
    InputMessageItem | InputFunctionCall | InputFunctionCallOutput | InputReasoning,
    Field(discriminator="type"),
]


# Legacy alias for backwards compatibility
class InputMessage(BaseModel):
    """Represents a single input message (legacy format)."""

    role: str = Field(description="Role of the message author.")
    content: str | list[ContentPart] = Field(
        description="Message content as raw text or structured content parts."
    )


class ResponseReasoning(BaseModel):
    effort: ReasoningEffort = Field(
        description="Controls the depth of the reasoning phase.",
    )

    @classmethod
    def validate_effort(cls, value: ReasoningEffort) -> ReasoningEffort:
        normalized = normalize_reasoning_effort(value)
        assert normalized is not None
        return normalized

    @property
    def normalized_effort(self) -> ReasoningEffort:
        return self.validate_effort(self.effort)


class ResponseRequest(BaseModel):
    """Defines the request schema for the /v1/responses endpoint."""

    model: str = Field(description="Model ID used to generate the response.")
    input: str | list[InputItem] = Field(
        description="Input as a string (shorthand for user message) or array of items."
    )
    stream: bool | None = Field(
        default=None,
        description="Whether to stream the response.",
    )
    truncation: Literal["auto", "disabled"] | None = Field(
        default=None,
        description="Controls whether server can truncate input when it exceeds context window.",
    )
    parallel_tool_calls: bool | None = Field(
        default=None,
        description="Whether to allow the model to run tool calls in parallel.",
    )
    max_tool_calls: int | None = Field(
        default=None,
        ge=1,
        description="Maximum number of tool calls the model can emit in one response.",
    )
    instructions: str | None = Field(
        default=None,
        description="System/developer instructions for the model.",
    )
    max_output_tokens: int | None = Field(
        default=None,
        ge=1,
        description="Upper bound for the number of tokens generated.",
    )
    temperature: float | None = Field(
        default=None,
        ge=0.0,
        le=2.0,
        description="Sampling temperature.",
    )
    top_p: float | None = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Nucleus sampling threshold.",
    )
    presence_penalty: float | None = Field(
        default=None,
        ge=-2.0,
        le=2.0,
        description="Penalizes new tokens based on whether they appear in the text so far.",
    )
    frequency_penalty: float | None = Field(
        default=None,
        ge=-2.0,
        le=2.0,
        description="Penalizes new tokens based on their frequency in the text so far.",
    )
    top_k: int | None = Field(
        default=None,
        ge=1,
        le=100,
        description="Controls the number of tokens considered at each step.",
    )
    min_p: float | None = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Minimum probability threshold for token consideration.",
    )
    top_logprobs: int | None = Field(
        default=None,
        ge=0,
        le=20,
        description="Number of top log probabilities to return per token.",
    )
    tool_choice: ToolUseMode | FunctionID = Field(
        default=ToolUseMode.AUTO,
        description="How the model should select which tool (or tools) to use when generating a response.",
    )
    tools: list[Function] | None = Field(
        default=None,
        description="A list of tools that the model can use to generate a response.",
    )
    text: ResponseFormat | None = Field(
        default=None,
        description="The format of the response.",
    )
    task: str | None = Field(
        default=None,
        description="Optional specialized task identifier for the decoder.",
    )
    reasoning: ResponseReasoning | bool | None = Field(
        default=None,
        description="Optional configuration for reasoning effort.",
    )
    include: list[str] | None = Field(
        default=None,
        description="What optional data to include in response (e.g., 'reasoning.encrypted_content', 'message.output_text.logprobs').",
    )
    metadata: dict[str, str] | None = Field(
        default=None,
        description="Key-value pairs attached to the response for tracking.",
    )
    # Persistence parameters - accepted but ignored (orchard-py is stateless)
    previous_response_id: str | None = Field(
        default=None,
        description="Ignored. Response chaining requires server-side storage.",
    )
    store: bool | None = Field(
        default=None,
        description="Ignored. Response persistence requires storage.",
    )
    background: bool | None = Field(
        default=None,
        description="Ignored. Async polling requires storage and task queue.",
    )

    @field_validator("reasoning", mode="before")
    @classmethod
    def _normalize_boolean_reasoning(cls, value: ResponseReasoning | bool | None):
        if value is None:
            return None
        if value is False:
            return None
        if value is True:
            return {"effort": DEFAULT_BOOLEAN_REASONING_EFFORT}
        return value

    @model_validator(mode="after")
    def _normalize_string_input(self) -> "ResponseRequest":
        """Convert string input to a single user message item."""
        if isinstance(self.input, str):
            object.__setattr__(
                self,
                "input",
                [InputMessageItem(role="user", content=self.input)],
            )
        return self

    def get_message_items(self) -> list[InputMessageItem]:
        """Convert input items into a uniform list of message items.

        InputFunctionCall → assistant message with tool_calls.
        InputFunctionCallOutput → tool message with call_id.
        InputReasoning → skipped (not representable as text).
        """
        if isinstance(self.input, str):
            return [InputMessageItem(role="user", content=self.input)]
        messages: list[InputMessageItem] = []
        for item in self.input:
            if isinstance(item, InputMessageItem):
                messages.append(item)
            elif isinstance(item, InputFunctionCall):
                messages.append(
                    InputMessageItem(
                        role="assistant",
                        content="",
                        tool_calls=[
                            ToolCall(
                                id=item.call_id,
                                function=ToolCallFunction(
                                    name=item.name,
                                    arguments=item.arguments,
                                ),
                            )
                        ],
                    )
                )
            elif isinstance(item, InputFunctionCallOutput):
                messages.append(
                    InputMessageItem(
                        role="tool",
                        content=item.output,
                        tool_call_id=item.call_id,
                    )
                )
        return messages



import httpx
import pytest

pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize("model_id", ["meta-llama/Llama-3.1-8B-Instruct", "moondream3"])
@pytest.mark.parametrize("batch_size", [2, 4, 8, 16])
async def test_multi_candidate_determinism(live_server, model_id, batch_size):
    """
    Tests a non-streaming request that requires multiple candidates to be generated.
    """
    server_url = live_server
    request_payload = {
        "model": model_id,
        "messages": [
            {
                "role": "user",
                "content": "Provide one friendly sentence introducing yourself.",
            }
        ],
        "max_completion_tokens": 64,
        "temperature": 0.0,  # Greedy for deterministic output
        "stream": False,
        "n": batch_size,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions", json=request_payload
        )

    assert response.status_code == 200
    response_data = response.json()

    assert len(response_data["choices"]) > 0
    # contents should be deterministic
    first_content = response_data["choices"][0]["message"]["content"]
    print(
        f"First content:\n{first_content}\n seen {len(response_data['choices'])} ✓",
        end="",
        flush=False,
    )

    drifts = 0
    for choice in response_data["choices"][1:]:
        assert choice["finish_reason"].lower() in ["length", "stop"]
        assert choice["message"]["content"] is not None
        assert len(choice["message"]["content"]) > 0
        if choice["message"]["content"] != first_content:
            drifts += 1
            print(f"Drift detected!:\n{choice['message']['content']}")
        else:
            print("✓", end="", flush=False)

    print()
    assert drifts == 0, f"Expected 0 drifts, got {drifts}"


@pytest.mark.parametrize("model_id", ["meta-llama/Llama-3.1-8B-Instruct", "moondream3"])
async def test_sequential_request_determinism(live_server, model_id):
    """
    Tests a sequential request that generates the same content multiple times.
    """
    server_url = live_server
    request_payload = {
        "model": model_id,
        "messages": [
            {
                "role": "user",
                "content": "Provide one friendly sentence introducing yourself.",
            }
        ],
        "max_completion_tokens": 64,
        "temperature": 0.0,  # Greedy for deterministic output
        "stream": False,
    }

    async def make_request():
        async with httpx.AsyncClient(timeout=30.0) as client:
            response = await client.post(
                f"{server_url}/v1/chat/completions", json=request_payload
            )
        return response

    first_response = None
    valid_responses = 0
    num_requests = 3
    for i in range(num_requests):
        response = await make_request()
        assert response.status_code == 200
        response_data = response.json()

        assert len(response_data["choices"]) > 0
        # contents should be deterministic
        content = response_data["choices"][0]["message"]["content"]
        if first_response is None:
            first_response = content
            valid_responses += 1
        elif first_response == content:
            valid_responses += 1
        else:
            print(f"Drift detected!:\n{content} at index {i}")
            break

    print(f"First response:\n{first_response}\n seen {valid_responses} ✓")
    for i in range(num_requests):
        if i == 0:
            print("✓", end="", flush=False)
        else:
            print(".", end="", flush=False)
    print()
    assert valid_responses == num_requests, (
        f"Expected {num_requests} valid responses, got {valid_responses}"
    )



import httpx
import pytest
from helpers import parse_sse_events

pytestmark = pytest.mark.asyncio

MODEL_ID = "meta-llama/Llama-3.1-8B-Instruct"


# ---------------------------------------------------------------------------
# Non-streaming text generation
# ---------------------------------------------------------------------------


async def test_responses_non_streaming_string_input(live_server):
    """String shorthand input produces a valid response."""
    payload = {
        "model": MODEL_ID,
        "input": "Say hello in one sentence.",
        "temperature": 0.0,
        "max_output_tokens": 32,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(f"{live_server}/v1/responses", json=payload)

    assert response.status_code == 200
    data = response.json()

    assert data["object"] == "response"
    assert data["id"].startswith("resp_")
    assert data["status"] == "completed"
    assert data["model"] == MODEL_ID
    assert isinstance(data["created_at"], int)

    assert isinstance(data["output"], list)
    assert len(data["output"]) >= 1
    msg = data["output"][0]
    assert msg["type"] == "message"
    assert msg["role"] == "assistant"
    assert len(msg["content"]) >= 1
    assert msg["content"][0]["type"] == "output_text"
    text = msg["content"][0]["text"]
    assert len(text) > 0
    print(f"Response: {text}")

    usage = data["usage"]
    assert usage["input_tokens"] > 0
    assert usage["output_tokens"] > 0
    assert usage["total_tokens"] == usage["input_tokens"] + usage["output_tokens"]


async def test_responses_non_streaming_message_items(live_server):
    """Array-of-items input produces a valid response."""
    payload = {
        "model": MODEL_ID,
        "input": [
            {
                "type": "message",
                "role": "user",
                "content": "What is 2+2? Answer with just the number.",
            }
        ],
        "temperature": 0.0,
        "max_output_tokens": 8,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(f"{live_server}/v1/responses", json=payload)

    assert response.status_code == 200
    data = response.json()
    assert data["status"] == "completed"

    text = data["output"][0]["content"][0]["text"]
    assert "4" in text
    print(f"Response: {text}")


async def test_responses_echo_fields(live_server):
    """Configuration fields are echoed back in the response."""
    payload = {
        "model": MODEL_ID,
        "input": "Hi",
        "temperature": 0.5,
        "top_p": 0.9,
        "max_output_tokens": 4,
        "metadata": {"test_key": "test_value"},
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(f"{live_server}/v1/responses", json=payload)

    assert response.status_code == 200
    data = response.json()

    assert data["temperature"] == 0.5
    assert data["top_p"] == 0.9
    assert data["metadata"] == {"test_key": "test_value"}


# ---------------------------------------------------------------------------
# Streaming text generation
# ---------------------------------------------------------------------------


async def test_responses_streaming_event_sequence(live_server):
    """Streaming produces the correct SSE event sequence."""
    payload = {
        "model": MODEL_ID,
        "input": "Say hello in one sentence.",
        "temperature": 0.0,
        "max_output_tokens": 32,
        "stream": True,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{live_server}/v1/responses",
            json=payload,
            headers={"Accept": "text/event-stream"},
        )

    assert response.status_code == 200
    events = parse_sse_events(response.text)
    event_types = [e["event"] for e in events]

    assert event_types[0] == "response.created"
    assert event_types[1] == "response.in_progress"
    assert "response.output_item.added" in event_types
    assert "response.content_part.added" in event_types
    assert "response.output_text.delta" in event_types
    assert "response.output_text.done" in event_types
    assert "response.content_part.done" in event_types
    assert "response.output_item.done" in event_types
    assert event_types[-2] == "response.completed"
    assert event_types[-1] == "done"

    seq_numbers = [
        e["data"]["sequence_number"]
        for e in events
        if isinstance(e["data"], dict) and "sequence_number" in e["data"]
    ]
    assert seq_numbers == sorted(seq_numbers)
    assert len(seq_numbers) == len(set(seq_numbers))


async def test_responses_streaming_delta_accumulation(live_server):
    """Accumulated deltas match the done event's full text."""
    payload = {
        "model": MODEL_ID,
        "input": "Count from 1 to 5.",
        "temperature": 0.0,
        "max_output_tokens": 64,
        "stream": True,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{live_server}/v1/responses",
            json=payload,
            headers={"Accept": "text/event-stream"},
        )

    events = parse_sse_events(response.text)

    accumulated = ""
    for e in events:
        if e["event"] == "response.output_text.delta":
            accumulated += e["data"]["delta"]

    done_events = [e for e in events if e["event"] == "response.output_text.done"]
    assert len(done_events) == 1
    assert accumulated == done_events[0]["data"]["text"]
    assert len(accumulated) > 0
    print(f"Streamed: {accumulated}")


async def test_responses_streaming_completed_snapshot(live_server):
    """The response.completed snapshot has correct status and usage."""
    payload = {
        "model": MODEL_ID,
        "input": "Hi",
        "temperature": 0.0,
        "max_output_tokens": 64,
        "stream": True,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{live_server}/v1/responses",
            json=payload,
            headers={"Accept": "text/event-stream"},
        )

    events = parse_sse_events(response.text)

    completed = [e for e in events if e["event"] == "response.completed"]
    assert len(completed) == 1
    snapshot = completed[0]["data"]["response"]
    assert snapshot["status"] == "completed"


# ---------------------------------------------------------------------------
# Incomplete response (max_output_tokens truncation)
# ---------------------------------------------------------------------------


async def test_responses_incomplete_non_streaming(live_server):
    """max_output_tokens=1 results in an incomplete response."""
    payload = {
        "model": MODEL_ID,
        "input": "Write a very long essay about the history of mathematics.",
        "temperature": 0.0,
        "max_output_tokens": 1,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(f"{live_server}/v1/responses", json=payload)

    assert response.status_code == 200
    data = response.json()

    assert data["status"] == "incomplete"
    assert data["incomplete_details"]["reason"] == "max_output_tokens"


async def test_responses_incomplete_streaming(live_server):
    """Streaming with max_output_tokens=1 emits response.incomplete."""
    payload = {
        "model": MODEL_ID,
        "input": "Write a very long essay about the history of mathematics.",
        "temperature": 0.0,
        "max_output_tokens": 1,
        "stream": True,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{live_server}/v1/responses",
            json=payload,
            headers={"Accept": "text/event-stream"},
        )

    events = parse_sse_events(response.text)
    event_types = [e["event"] for e in events]

    assert "response.incomplete" in event_types
    incomplete = [e for e in events if e["event"] == "response.incomplete"]
    snapshot = incomplete[0]["data"]["response"]
    assert snapshot["status"] == "incomplete"
    assert snapshot["incomplete_details"]["reason"] == "max_output_tokens"
    assert event_types[-1] == "done"


# ---------------------------------------------------------------------------
# Instructions (system prompt)
# ---------------------------------------------------------------------------


async def test_responses_instructions(live_server):
    """The instructions field works as a system prompt."""
    payload = {
        "model": MODEL_ID,
        "input": "What is your name?",
        "instructions": "You are a helpful assistant named Orchard. Always introduce yourself by name.",
        "temperature": 0.0,
        "max_output_tokens": 64,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(f"{live_server}/v1/responses", json=payload)

    assert response.status_code == 200
    data = response.json()

    text = data["output"][0]["content"][0]["text"].lower()
    assert "orchard" in text
    print(f"Response: {text}")



import httpx
import pytest

pytestmark = pytest.mark.asyncio

MODEL_ID = "moondream3"


async def test_chat_completion_with_logprobs(live_server):
    """
    Tests a chat completion request with logprobs enabled.
    Verifies that the system can return top log probabilities for each generated token.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [{"role": "user", "content": "Say hello"}],
        "max_completion_tokens": 3,  # Generate a few tokens
        "temperature": 1.0,
        "logprobs": True,  # Enable logprobs
        "top_logprobs": 5,  # Request top 5 log probabilities
        "stream": False,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions", json=request_payload
        )

    # --- Assertions ---
    assert response.status_code == 200
    response_data = response.json()

    assert "choices" in response_data
    assert len(response_data["choices"]) > 0

    choice = response_data["choices"][0]
    assert "logprobs" in choice

    # If logprobs were requested and generated, they should be present
    if choice["logprobs"] is not None:
        logprobs_data = choice["logprobs"]
        assert "content" in logprobs_data
        assert isinstance(logprobs_data["content"], list)

        # Check that we have logprobs for each generated token
        assert len(logprobs_data["content"]) > 0

        # Verify the structure of each logprob entry
        for token_logprob in logprobs_data["content"]:
            assert isinstance(token_logprob, dict)
            assert "token" in token_logprob
            assert "logprob" in token_logprob
            assert isinstance(token_logprob["logprob"], int | float)

            # Check top_logprobs if present
            if token_logprob.get("top_logprobs"):
                assert isinstance(token_logprob["top_logprobs"], list)
                assert (
                    len(token_logprob["top_logprobs"]) <= 5
                )  # Should not exceed requested amount

                # Verify each top logprob entry
                for top_logprob in token_logprob["top_logprobs"]:
                    assert "token" in top_logprob
                    assert "logprob" in top_logprob
                    assert isinstance(top_logprob["logprob"], int | float)

                # Verify logprobs are sorted in descending order
                logprob_values = [
                    item["logprob"] for item in token_logprob["top_logprobs"]
                ]
                assert logprob_values == sorted(logprob_values, reverse=True)


async def test_chat_completion_without_logprobs(live_server):
    """
    Tests that when logprobs is not requested, they are not included in the response.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [{"role": "user", "content": "Say hello"}],
        "max_completion_tokens": 3,
        "temperature": 1.0,
        "logprobs": False,  # Explicitly disable logprobs
        "stream": False,
    }

    async with httpx.AsyncClient(timeout=30.0) as client:
        response = await client.post(
            f"{server_url}/v1/chat/completions", json=request_payload
        )

    assert response.status_code == 200
    response_data = response.json()

    choice = response_data["choices"][0]
    # When logprobs is false, the field should be None or absent
    assert choice.get("logprobs") is None


async def test_chat_completion_logprobs_streaming(live_server):
    """
    Tests that logprobs work correctly with streaming responses.
    """
    server_url = live_server
    request_payload = {
        "model": MODEL_ID,
        "messages": [{"role": "user", "content": "Count to three"}],
        "max_completion_tokens": 5,
        "temperature": 1.0,
        "logprobs": True,
        "top_logprobs": 3,
        "stream": True,  # Enable streaming
    }

    chunks = []
    async with httpx.AsyncClient(timeout=30.0) as client:
        async with client.stream(
            "POST", f"{server_url}/v1/chat/completions", json=request_payload
        ) as response:
            assert response.status_code == 200
            async for line in response.aiter_lines():
                if line.startswith("data: "):
                    data_str = line[6:]  # Remove "data: " prefix
                    if data_str == "[DONE]":
                        break
                    chunks.append(data_str)

    # Verify we received multiple chunks
    assert len(chunks) > 0

    # Parse and verify chunk structure
    import json

    for chunk_str in chunks:
        if chunk_str.strip():
            chunk_data = json.loads(chunk_str)
            assert "choices" in chunk_data

            # Check if this chunk has logprobs (not all chunks might have them)
            choice = chunk_data["choices"][0]
            if "logprobs" in choice and choice["logprobs"] is not None:
                # Streaming logprobs might have a different structure
                # Adjust assertions based on actual implementation
                pass



from __future__ import annotations

import secrets
import time
from enum import Enum
from typing import Literal

from pydantic import BaseModel, Field

from orchard.server.models.responses.format import ResponseFormat
from orchard.server.models.responses.tools import (
    Function,
    FunctionID,
    ToolUseMode,
)
from orchard.server.models.tools import generate_tool_call_id

# --- Constants ---
RESPONSE_ID_PREFIX = "resp_"
RESPONSE_OBJECT = "response"
MESSAGE_ID_PREFIX = "msg_"
MESSAGE_OBJECT = "message"
OUTPUT_TEXT_OBJECT = "output_text"
FUNCTION_CALL_ID_PREFIX = "fc_"


def generate_message_id(prefix: str = MESSAGE_ID_PREFIX) -> str:
    random_part = secrets.token_urlsafe(22)
    return f"{prefix}{random_part}"


def generate_response_id(prefix: str = RESPONSE_ID_PREFIX) -> str:
    random_part = secrets.token_urlsafe(22)
    return f"{prefix}{random_part}"


def generate_function_call_id(prefix: str = FUNCTION_CALL_ID_PREFIX) -> str:
    random_part = secrets.token_urlsafe(22)
    return f"{prefix}{random_part}"


def get_current_timestamp() -> int:
    return int(time.time())


class OutputTextContent(BaseModel):
    """Represents text content within an output message."""

    type: Literal["output_text"] = OUTPUT_TEXT_OBJECT
    text: str = Field(description="The generated text content.")
    annotations: list = Field(
        default_factory=list,
        description="Annotations for the text content.",
    )
    logprobs: list = Field(
        default_factory=list,
        description="Log probability information for the text.",
    )


class OutputStatus(Enum):
    """Represents the status of an output item."""

    COMPLETED = "completed"
    INCOMPLETE = "incomplete"
    IN_PROGRESS = "in_progress"
    FAILED = "failed"


class OutputMessage(BaseModel):
    """Represents an assistant message output item."""

    type: Literal["message"] = "message"
    id: str = Field(
        default_factory=generate_message_id,
        description="Unique message identifier.",
    )
    status: OutputStatus = Field(
        default=OutputStatus.COMPLETED,
        description="The status of the message.",
    )
    role: Literal["assistant"] = "assistant"
    content: list[OutputTextContent] = Field(
        default_factory=list,
        description="Content parts generated by the assistant.",
    )


class OutputFunctionCall(BaseModel):
    """Represents a function call output item."""

    type: Literal["function_call"] = "function_call"
    id: str = Field(
        default_factory=generate_function_call_id,
        description="Unique function call item identifier.",
    )
    call_id: str = Field(
        default_factory=generate_tool_call_id,
        description="Identifier used to match function call with its output.",
    )
    name: str = Field(description="The name of the function to call.")
    arguments: str = Field(description="JSON-encoded arguments for the function.")
    status: OutputStatus = Field(
        default=OutputStatus.COMPLETED,
        description="The status of the function call.",
    )


class ReasoningSummaryTextContent(BaseModel):
    """Represents a summary text within reasoning output."""

    type: Literal["summary_text"] = "summary_text"
    text: str = Field(description="Summary text of the reasoning.")


class ReasoningContent(BaseModel):
    """Represents reasoning text content within reasoning output."""

    type: Literal["reasoning_text"] = "reasoning_text"
    text: str = Field(description="The reasoning/thinking content.")


class OutputReasoning(BaseModel):
    """Represents reasoning output item."""

    type: Literal["reasoning"] = "reasoning"
    id: str = Field(
        default_factory=lambda: generate_message_id("reasoning_"),
        description="Unique reasoning item identifier.",
    )
    status: OutputStatus = Field(
        default=OutputStatus.COMPLETED,
        description="The status of the reasoning.",
    )
    summary: list[ReasoningSummaryTextContent] = Field(
        default_factory=list,
        description="Summary of the reasoning.",
    )
    content: list[ReasoningContent] = Field(
        default_factory=list,
        description="The full reasoning content.",
    )
    encrypted_content: str | None = Field(
        default=None,
        description="Encrypted reasoning for continuation.",
    )


OutputItem = OutputMessage | OutputFunctionCall | OutputReasoning


class IncompleteDetails(BaseModel):
    """Details about why a response was incomplete."""

    reason: str = Field(
        description="The reason the response was incomplete (e.g., 'max_output_tokens', 'content_filter').",
    )


class ResponseError(BaseModel):
    """Structured error information when response fails."""

    code: str = Field(
        description="Error code (e.g., 'server_error', 'invalid_request')."
    )
    message: str = Field(description="Human-readable error message.")


class InputTokensDetails(BaseModel):
    """Details about input token usage."""

    cached_tokens: int = Field(
        default=0,
        description="Number of tokens served from cache.",
    )


class OutputTokensDetails(BaseModel):
    """Details about output token usage."""

    reasoning_tokens: int = Field(
        default=0,
        description="Number of tokens used for reasoning/thinking.",
    )


class ResponseUsage(BaseModel):
    """Provides token usage statistics for the chat completion request."""

    input_tokens: int = Field(
        description="The number of tokens constituting the input prompt(s)."
    )
    output_tokens: int = Field(
        description="The total number of tokens generated across all completion choices."
    )
    total_tokens: int = Field(
        description="The sum of `input_tokens` and `output_tokens`."
    )
    input_tokens_details: InputTokensDetails | None = Field(
        default=None,
        description="Breakdown of input token usage.",
    )
    output_tokens_details: OutputTokensDetails | None = Field(
        default=None,
        description="Breakdown of output token usage.",
    )


class ResponseObject(BaseModel):
    """Defines the response schema for the /v1/responses endpoint."""

    id: str = Field(
        default_factory=generate_response_id,
        description="Unique identifier for the response.",
    )
    object: Literal["response"] = RESPONSE_OBJECT
    created_at: int = Field(
        default_factory=get_current_timestamp,
        description="Unix timestamp of response creation.",
    )
    completed_at: int | None = Field(
        default=None,
        description="Unix timestamp when response generation completed.",
    )
    status: OutputStatus = Field(
        default=OutputStatus.COMPLETED,
        description="The status of the response.",
    )
    incomplete_details: IncompleteDetails | None = Field(
        default=None,
        description="Details about why the response was incomplete, if applicable.",
    )
    error: ResponseError | None = Field(
        default=None,
        description="Error information if the response failed.",
    )
    model: str = Field(description="Model ID used for the response.")
    output: list[OutputItem] = Field(
        default_factory=list,
        description="List of output items (messages, function calls, reasoning).",
    )
    usage: ResponseUsage | None = Field(
        default=None,
        description="Usage statistics for the response request.",
    )
    # Echoed configuration
    metadata: dict[str, str] | None = Field(
        default=None,
        description="Key-value pairs attached to the response.",
    )
    parallel_tool_calls: bool = Field(
        default=False,
        description="Whether parallel tool calls were enabled.",
    )
    temperature: float | None = Field(
        default=None,
        description="Sampling temperature used.",
    )
    top_p: float | None = Field(
        default=None,
        description="Nucleus sampling threshold used.",
    )
    presence_penalty: float | None = Field(
        default=None,
        description="Presence penalty used.",
    )
    frequency_penalty: float | None = Field(
        default=None,
        description="Frequency penalty used.",
    )
    top_k: int | None = Field(
        default=None,
        description="Top-k value used.",
    )
    min_p: float | None = Field(
        default=None,
        description="Min-p value used.",
    )
    truncation: str | None = Field(
        default=None,
        description="How input was truncated.",
    )
    instructions: str | None = Field(
        default=None,
        description="System/developer instructions used.",
    )
    max_output_tokens: int | None = Field(
        default=None,
        description="Maximum output tokens allowed.",
    )
    top_logprobs: int | None = Field(
        default=None,
        description="Number of top log probabilities returned.",
    )
    tool_choice: ToolUseMode | FunctionID = Field(
        default=ToolUseMode.AUTO,
        description="Tool choice mode used.",
    )
    tools: list[Function] = Field(
        default_factory=list,
        description="Tools that were available.",
    )
    max_tool_calls: int | None = Field(
        default=None,
        description="Maximum number of tool calls allowed.",
    )
    text: ResponseFormat | None = Field(
        default=None,
        description="Response format used.",
    )
    reasoning: dict | None = Field(
        default=None,
        description="Reasoning configuration used.",
    )

    @property
    def output_text(self) -> str:
        """Concatenate all assistant output text parts into a single string."""
        chunks: list[str] = []
        for item in self.output:
            if isinstance(item, OutputMessage):
                for part in item.content:
                    chunks.append(part.text)
        return "".join(chunks)

    @property
    def tool_calls(self) -> list[OutputFunctionCall]:
        """Return all function-call output items in this response."""
        return [
            item for item in self.output if isinstance(item, OutputFunctionCall)
        ]



"""CLI helper for generating preload manifest files shared with the C++ engine."""

from __future__ import annotations

import argparse
import json
import sys
from collections.abc import Iterable
from dataclasses import asdict, dataclass
from pathlib import Path

from orchard.app.model_resolver import (
    ModelResolutionError,
    ModelResolver,
)


@dataclass(slots=True, frozen=True)
class PreloadEntry:
    requested_id: str
    canonical_id: str
    model_path: str


def resolve_models(models: Iterable[str]) -> list[PreloadEntry]:
    """Resolve model identifiers to local paths via HuggingFace."""
    resolver = ModelResolver()
    entries: list[PreloadEntry] = []
    seen: set[tuple[str, str]] = set()

    for requested in models:
        try:
            resolved = resolver.resolve(requested)
        except ModelResolutionError as exc:  # pragma: no cover - defensive logging path
            raise SystemExit(
                f"Failed to resolve model '{requested}': {exc} (candidates={exc.candidates})"
            ) from exc
        entry_key = (resolved.canonical_id.lower(), str(resolved.model_path))
        if entry_key in seen:
            continue
        seen.add(entry_key)
        entries.append(
            PreloadEntry(
                requested_id=requested,
                canonical_id=resolved.canonical_id,
                model_path=str(resolved.model_path),
            )
        )
    return entries


def write_manifest(entries: list[PreloadEntry], output_path: Path) -> None:
    manifest = {"models": [asdict(entry) for entry in entries]}
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(json.dumps(manifest, indent=2), encoding="utf-8")


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Generate a preload manifest consumed by orchard engine."
    )
    parser.add_argument(
        "--output",
        required=True,
        type=Path,
        help="Destination path for the manifest JSON file.",
    )
    parser.add_argument(
        "--model",
        dest="models",
        action="append",
        default=[],
        help="Model identifier (HF repo ID, alias, or local path). Repeat for multiple models.",
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Suppress informational stdout output.",
    )
    return parser.parse_args(argv)


def main(argv: list[str] | None = None) -> int:
    args = parse_args(argv)
    if not args.models:
        raise SystemExit("At least one --model must be provided.")

    entries = resolve_models(args.models)
    write_manifest(entries, args.output)

    if not args.quiet:
        print(f"Wrote manifest with {len(entries)} model(s) to {args.output}")
        for entry in entries:
            print(
                f"  {entry.requested_id} -> {entry.canonical_id} ({entry.model_path})"
            )

    return 0


if __name__ == "__main__":
    sys.exit(main())



import json
import logging
from pathlib import Path
from typing import Any

import yaml
from jinja2 import Environment, FileSystemLoader

from orchard.formatter.control_tokens import ControlTokens, load_control_tokens

logger = logging.getLogger(__name__)


def determine_model_type(config: dict) -> str:
    """Determine the model type from the model path."""
    model_type = config.get("model_type", "llama")
    if model_type == "llama" or model_type == "llama3":
        return "llama3"

    if model_type == "moondream3" or model_type == "moondream":
        return "moondream3"

    return model_type


# from https://github.com/huggingface/transformers/blob/7769f660935b5d48b73bf6711d0a78b6f8f98739/src/transformers/utils/chat_template_utils.py#L447C1-L451C1
def tojson(x, ensure_ascii=False, indent=None, separators=None, sort_keys=False):
    # We override the built-in tojson filter because Jinja's default filter escapes HTML characters
    # We also expose some options like custom indents and separators
    return json.dumps(
        x,
        ensure_ascii=ensure_ascii,
        indent=indent,
        separators=separators,
        sort_keys=sort_keys,
    )


class ChatFormatter:
    """
    Handles the application of chat templates to conversation histories.
    """

    @property
    def image_placeholder(self) -> str:
        """Resolve the image placeholder token from capabilities or control tokens."""
        vision = self.capabilities.get("vision", {})
        # Explicit placeholder (e.g. moondream: vision.placeholders.image)
        placeholder = vision.get("placeholders", {}).get("image")
        if placeholder:
            return placeholder
        # Vision start token (e.g. gemma: vision.tokens.start)
        start = vision.get("tokens", {}).get("start")
        if start:
            return start
        # Legacy fallback to control_tokens
        if self.control_tokens.start_image_token:
            return self.control_tokens.start_image_token
        return "<|image|>"

    @property
    def should_clip_image_placeholder(self) -> bool:
        """Whether to clip the image placeholder from text segments.

        If the placeholder is a real token (has a start_image_token or vision.tokens.start),
        it stays in the text for tokenization. If it's a synthetic placeholder
        (like <|image|>), it gets clipped.
        """
        vision = self.capabilities.get("vision", {})
        has_vision_tokens = bool(vision.get("tokens", {}).get("start"))
        has_start_token = bool(self.control_tokens.start_image_token)
        return not (has_vision_tokens or has_start_token)

    def __init__(self, model_path: str):
        self.model_path = Path(model_path)

        # 1. Load tokenizer_config to determine model family
        tokenizer_config_path = self.model_path / "config.json"
        if not tokenizer_config_path.exists():
            raise FileNotFoundError(
                f"tokenizer_config.json not found in {self.model_path}"
            )
        with open(tokenizer_config_path) as f:
            self.tokenizer_config = json.load(f)

        model_type = determine_model_type(self.tokenizer_config)
        profile_dir = Path(__file__).parent / "profiles" / model_type
        if not profile_dir.is_dir():
            raise ValueError(
                f"Profile directory for model_type '{model_type}' not found at "
                f"{profile_dir}"
            )
        self.profile_dir = profile_dir

        self.control_tokens: ControlTokens = load_control_tokens(profile_dir)

        # 2b. Load capabilities manifest
        caps_path = profile_dir / "capabilities.yaml"
        self.capabilities: dict[str, Any] = (
            yaml.safe_load(caps_path.read_text()) if caps_path.exists() else {}
        )

        # 3. Set up Jinja2 environment
        self.jinja_env = Environment(
            loader=FileSystemLoader(profile_dir), trim_blocks=True, lstrip_blocks=True
        )
        self.jinja_env.filters["tojson"] = tojson
        self.template = self.jinja_env.get_template("chat_template.jinja")

    def apply_template(
        self,
        conversation: list[dict[str, Any]],
        add_generation_prompt: bool = True,
        reasoning: bool = False,
        task: str | None = None,
        prefill: str | None = None,
        tools: list[dict[str, Any]] | None = None,
    ) -> str:
        """
        Applies the loaded chat template to a conversation.

        Args:
            conversation: A list of message dictionaries, e.g., [{"role": "user", "content": "..."}].
            add_generation_prompt: Whether to add the assistant prompt turn.
            reasoning: Whether to add the conditional reasoning prompt logic.
            task: Optional task name for task-specific formatting (e.g., "caption_normal", "detect").
            prefill: Optional string to prefill the assistant response with.
            tools: Optional list of tool schema dicts to render inline.
        Returns:
            A single, fully formatted string ready for tokenization.
        """
        context = {
            "interactions": conversation,
            "add_generation_prompt": add_generation_prompt,
            "begin_of_text": self.control_tokens.begin_of_text,
            "end_of_sequence": self.control_tokens.end_of_sequence,
            "end_of_message": self.control_tokens.end_of_message,
            "start_image_token": self.control_tokens.start_image_token,
            "end_image_token": self.control_tokens.end_image_token,
            "reasoning": reasoning,
            "task": task,
            "roles": self.control_tokens.roles.model_dump(),
            "prefill": prefill,
            "capabilities": self.capabilities,
            "tools": tools,
        }
        return self.template.render(**context)

    def get_coord_placeholder(self) -> str | None:
        """Extract coord placeholder from capabilities.yaml (pointing or gaze_detection)."""
        for cap_name in ("pointing", "gaze_detection"):
            cap = self.capabilities.get(cap_name, {})
            placeholder = cap.get("placeholders", {}).get("coord")
            if placeholder:
                return placeholder
        return None

    def get_tool_calling_tokens(self) -> dict[str, Any]:
        """Extract tool calling delimiter tokens from capabilities.yaml."""
        tool_caps = self.capabilities.get("tool_calling", {})
        formats = tool_caps.get("formats") or []

        serialized_formats: list[dict[str, str]] = []
        section_start = ""
        section_end = ""

        for index, fmt in enumerate(formats):
            tokens = fmt.get("tokens", {})
            serialized_formats.append(
                {
                    "call_start": tokens.get("start", ""),
                    "call_end": tokens.get("end", ""),
                }
            )
            if index == 0:
                section_start = tokens.get("section_start", "")
                section_end = tokens.get("section_end", "")

        return {
            "formats": serialized_formats,
            "section_start": section_start,
            "section_end": section_end,
        }



import argparse
import asyncio
import logging
import sys
from pathlib import Path

import uvicorn

from orchard.engine.fetch import (
    ORCHARD_HOME,
    check_for_updates,
    download_engine,
    get_installed_version,
)
from orchard.engine.inference_engine import InferenceEngine
from orchard.server.app import create_app

# Configure logging for the CLI
logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s [%(levelname)s] %(message)s"
)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("httpx").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


async def run_serve(args: argparse.Namespace):
    """Handler for the 'serve' command."""
    logger.debug("Initializing local inference engine...")
    engine_log_path = Path(args.engine_log_file) if args.engine_log_file else None
    async with InferenceEngine(
        engine_log_file=engine_log_path,
        startup_timeout=args.engine_startup_timeout,
    ) as local_engine:
        logger.debug("Engine started.")
        if args.models:
            await local_engine.load_models(args.models)

        fastapi_app = create_app(local_engine)
        config = uvicorn.Config(
            fastapi_app,
            host=args.host,
            port=args.port,
            log_level="warning",
            loop="asyncio",
        )
        server = uvicorn.Server(config)
        await server.serve()

    logger.debug("Server has shut down.")


def run_engine_stop(args: argparse.Namespace):
    """Handler for the 'engine stop' command."""
    logger.info("Attempting to shut down the shared orchard engine process...")
    if InferenceEngine.shutdown(timeout=args.timeout):
        logger.info("Engine shutdown successful.")
    else:
        logger.warning("Engine shutdown was forceful or timed out.")
        sys.exit(1)


def run_upgrade(args: argparse.Namespace):
    """Handler for the 'upgrade' command."""
    channel = args.channel

    if args.reinstall:
        print(f"\033[34m→\033[0m Reinstalling \033[32m{channel}\033[0m build...")
        # Clear existing installation
        bin_dir = ORCHARD_HOME / "bin"
        version_file = ORCHARD_HOME / "version.txt"
        if bin_dir.exists():
            import shutil

            shutil.rmtree(bin_dir)
            print("\033[34m→\033[0m Cleared existing installation")
        if version_file.exists():
            version_file.unlink()
    else:
        print(f"\033[34m→\033[0m Upgrading to latest \033[32m{channel}\033[0m build...")

    if current := get_installed_version():
        print(f"\033[34m→\033[0m Current version: \033[34m{current}\033[0m")

    latest_version = check_for_updates(channel=channel)
    if latest_version:
        print(f"\033[34m→\033[0m Latest version: \033[34m{latest_version}\033[0m")
    else:
        latest_version = current

    try:
        if latest_version == current and not args.reinstall:
            print(
                f"\033[32m✓\033[0m Already on latest: \033[34m{latest_version}\033[0m"
            )
        else:
            download_engine(channel=channel)
    except Exception as e:
        print(f"\033[31m✗\033[0m Upgrade failed: \033[31m{e}\033[0m")
        sys.exit(1)


def main():
    """Main CLI entry point for the Proxy Inference Engine."""
    parser = argparse.ArgumentParser(
        description="Proxy Inference Engine (PIE) command-line interface."
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    # --- 'serve' command ---
    serve_parser = subparsers.add_parser("serve", help="Start the FastAPI server.")
    serve_parser.add_argument(
        "--engine-log-file", type=str, default=None, help="Path to the engine log file."
    )
    serve_parser.add_argument(
        "--host", type=str, default="127.0.0.1", help="Host to bind the server to."
    )
    serve_parser.add_argument(
        "--port", type=int, default=8000, help="Port to listen on."
    )
    serve_parser.add_argument(
        "--model",
        "--models",
        dest="models",
        nargs="*",
        help="Model identifier(s) to preload on startup.",
    )
    serve_parser.add_argument(
        "--engine-startup-timeout",
        type=float,
        default=60.0,
        help="Seconds to wait for the C++ engine to start.",
    )
    serve_parser.set_defaults(func=run_serve)

    # --- 'upgrade' command ---
    upgrade_parser = subparsers.add_parser(
        "upgrade", help="Download and install the latest engine binary."
    )
    upgrade_parser.add_argument(
        "channel",
        nargs="?",
        default="stable",
        help="Release channel to pull from (default: stable).",
    )
    upgrade_parser.add_argument(
        "--reinstall",
        action="store_true",
        help="Force reinstall by clearing existing installation first.",
    )
    upgrade_parser.set_defaults(func=run_upgrade)

    # --- 'engine' command group ---
    engine_parser = subparsers.add_parser(
        "engine", help="Manage the background engine process."
    )
    engine_subparsers = engine_parser.add_subparsers(
        dest="engine_command", required=True
    )

    # --- 'engine stop' command ---
    stop_parser = engine_subparsers.add_parser(
        "stop", help="Stop the shared orchard engine process."
    )
    stop_parser.add_argument(
        "--timeout",
        type=float,
        default=15.0,
        help="Seconds to wait for graceful shutdown.",
    )
    stop_parser.set_defaults(func=run_engine_stop)

    args = parser.parse_args()
    try:
        if asyncio.iscoroutinefunction(args.func):
            asyncio.run(args.func(args))
        else:
            args.func(args)
    except KeyboardInterrupt:
        logger.debug("Keyboard interrupt received.")
        sys.exit(0)
    except Exception as e:
        logger.error("Error: %s", e)
        sys.exit(1)


if __name__ == "__main__":
    main()
